# -*- coding: utf-8 -*-
import Cookie
import logging
import datetime
import hashlib
import threading
from sqlalchemy.exc import StatementError

from ckanext.dgu.drupalclient import DrupalClient, DrupalXmlRpcSetupError, \
     DrupalRequestError
from xmlrpclib import ServerProxy

log = logging.getLogger(__name__)

class DrupalAuthMiddleware(object):
    '''Allows CKAN user to login via Drupal. It looks for the Drupal cookie
    and gets user details from Drupal using XMLRPC.
    so works side-by-side with normal CKAN logins.'''
    
    #Roles that CKAN will get from Drupal and should copy them in its groups
    SYNCED_ROLES = [{'name':'glavni_uredniki', 'title':u'Glavni uredniki', 'drupal_name':u'Glavni urednik'},
                    {'name':'podrocni_uredniki', 'title':u'Področni uredniki', 'drupal_name':u'Področni urednik'},
                    {'name':'vsebinski_uredniki', 'title':u'Vsebinski uredniki', 'drupal_name':u'Vsebinski urednik'},
                    #{'name':'tehnicni_uredniki', 'title':u'Tehnični uredniki', 'drupal_name':u'Tehnični urednik'},
                    {'name':'sistemski_administratorji', 'title':u'Sistemski administratorji', 'drupal_name':u'Sistemski administrator'}]

    def _init_translator(self):
        from ckanext.dgu.bin import common
        common.register_translator()

    def _init_translator_and_roles(self):
        # MK: I introduced the below when creation of new user roles caused translation errors.
        # Lately, it started causing problems when 404 error should be displayed (they were displayed as
        # separate pages instead of in CKAN. I only call it when catching exception now.
        #_init_translator(self)
    
        # Publisher roles
        self._init_roles()

    def __init__(self, app, app_conf):
        self.app = app
        # MK: DrupalClient creates an instance of ServerProxy which causes errors when multiple threads
        # want to use its connection at the same time, so we create a separate instance for each thread
        self.localthread = threading.local()
        self.localthread.drupal_client = None
        self._user_name_prefix = 'user_d'
        self._drupal_session_prefix = 'drupal_'

        minutes_between_checking_drupal_cookie = (app_conf.get('minutes_between_checking_drupal_cookie') if app_conf else None) or 30
        self.seconds_between_checking_drupal_cookie = int(minutes_between_checking_drupal_cookie) * 60
        # if that int() raises a ValueError then the app will not start
        self._init_translator_and_roles()
        
    # removes unneeded cookies in case they would cause parsing errors
    def _clean_cookie_string(self, cookie_string):
        return ';'.join([cookie for cookie in cookie_string.split(';') if cookie.strip().startswith('auth_tkt=') or cookie.strip().startswith('SESS') or cookie.strip().startswith('SSESS')])

    def _parse_cookies(self, environ):
        is_ckan_cookie = [False]
        drupal_session_id = [False]
        from pylons import config
        server_name=config['ckan.site_url']
        server_name=server_name.replace('http://','').replace('https://','')
        for k, v in environ.items():
            key = k.lower()
            if key == 'http_cookie':
                cleaned_cookies = self._clean_cookie_string(v)
                is_ckan_cookie[0] = self._is_this_a_ckan_cookie(cleaned_cookies)
                drupal_session_id[0] = self._drupal_cookie_parse(cleaned_cookies, server_name)
        is_ckan_cookie = is_ckan_cookie[0]
        drupal_session_id = drupal_session_id[0]
        return is_ckan_cookie, drupal_session_id

    @staticmethod
    def _drupal_cookie_parse(cookie_string, server_name):
        '''Returns the Drupal Session ID from the cookie string.'''
        import xml.sax.saxutils
        cookies = Cookie.SimpleCookie()
        try:
            cookies.load(str(cookie_string))
        except Cookie.CookieError:
            log.error("Received invalid cookie: %s" % cookie_string)
            return False
        similar_cookies = []
        for cookie in cookies:
            if cookie.startswith('SESS') or cookie.startswith('SSESS'):
                # Drupal 6 uses md5, Drupal 7 uses sha256
                server_hash = hashlib.sha256(server_name).hexdigest()[:32]
                if cookie == 'SESS%s' % server_hash:
                    log.debug('Drupal cookie found for server request %s', server_name)
                    return xml.sax.saxutils.escape(cookies[cookie].value)
                elif cookie == 'SSESS%s' % server_hash:
                    log.debug('Drupal cookie (secure) found for server request %s', server_name)
                    return xml.sax.saxutils.escape(cookies[cookie].value)
                else:
                    similar_cookies.append(cookie)
        if similar_cookies:
            log.debug('Drupal cookies ignored with incorrect hash for server %r: %r',
                      server_name, similar_cookies)
        return None
    
    def _init_roles(self):
        import ckan.plugins.toolkit as tk
        import ckan.model as model
        from ckan.model.meta import Session
        
        context = {}
        data_dict = {}
        needs_sync = False
        data_dict['groups'] = [g['name'] for g in self.SYNCED_ROLES]
        context['user'] = 'frontend'
        retrieved_list=[]
        try:
            retrieved_list = tk.get_action('group_list')(context=context, data_dict=data_dict)
        except StatementError as se:
            log.warning("Statement error in drupal_auth, removing Session to reset it")
            log.debug("Full error: "+str(se))
            Session.remove()
            try:
                retrieved_list = tk.get_action('group_list')(context=context, data_dict=data_dict)
            except StatementError as e:
                log.error('Statement error even after dropping Session and retrying: '+str(e))
                raise StatementError("Statement error even after dropping Session and retrying: "+str(e.message), e.statement, e.params, e.orig)
                
        for g in data_dict['groups']:
            if g not in retrieved_list:
                needs_sync = True
                break
            
        if needs_sync:
            rev = model.repo.new_revision()
            rev.author = u'vs_admin'
            rev.message = u'Initializing groups for VS'
            
            for g in self.SYNCED_ROLES:
                if g not in retrieved_list:
                    new_group = model.Group(name=g['name'])
                    setattr(new_group, 'title', g['title'])
                    setattr(new_group, 'type', 'VS_role')
                    Session.add(new_group) 
            Session.commit()
            Session.close()
            
        if not self.SYNCED_ROLES[0].has_key('id'): #resync IDs
            data_dict['all_fields']=True
            retrieved_list = tk.get_action('group_list')(context=context, data_dict=data_dict)
            for g in retrieved_list:
                for s in self.SYNCED_ROLES:
                    if s['name'] == g['name']:
                        s['id'] = g['id']
                        break

    @staticmethod
    def _is_this_a_ckan_cookie(cookie_string):
        cookies = Cookie.SimpleCookie()
        try:
            cookies.load(str(cookie_string))
        except Cookie.CookieError:
            log.warning("Received invalid cookie: %s" % cookie_string)
            return False

        if not 'auth_tkt' in cookies:
            return False
        return True

    def _munge_drupal_id_to_ckan_user_name(self, drupal_id):
        drupal_id.lower().replace(' ', '_')
        return u'%s%s' % (self._user_name_prefix, drupal_id)

    def _log_out(self, environ, new_headers):
        # don't progress the user info for this request
        environ['REMOTE_USER'] = None
        environ['repoze.who.identity'] = None
        # tell auth_tkt to logout whilst adding the header to tell
        # the browser to delete the cookie
        identity = {}
        headers = environ['repoze.who.plugins']['dgu_auth_tkt'].forget(environ, identity)
        if headers:
            new_headers.extend(headers)
        # Remove cookie from request, so that if we are doing a login again in this request then
        # it is aware of the cookie removal
        #log.debug('Removing cookies from request: %r', environ.get('HTTP_COOKIE', ''))
        cookies = environ.get('HTTP_COOKIE', '').split('; ')
        cookies = '; '.join([cookie for cookie in cookies if not cookie.startswith('auth_tkt=')])
        environ['HTTP_COOKIE'] = cookies
        #log.debug('Cookies in request now: %r', environ['HTTP_COOKIE'])

        log.debug('Logged out Drupal user')

    def __call__(self, environ, start_response):
        '''
        Middleware that sets the CKAN logged-in/logged-out status according
        to Drupal's logged-in/logged-out status.

        Every request comes through here before hitting CKAN because it is
        configured as middleware.
        '''
        new_headers = []
        self._init_translator_and_roles()
        self.do_drupal_login_logout(environ, new_headers)

        #log.debug('New headers: %r', new_headers)
        def cookie_setting_start_response(status, headers, exc_info=None):
            if headers:
                headers.extend(new_headers)
            else:
                headers = new_headers
            return start_response(status, headers, exc_info)
        new_start_response = cookie_setting_start_response

        return self.app(environ, new_start_response)

    def do_drupal_login_logout(self, environ, new_headers):
        '''Looks at cookies and auth_tkt and may tell auth_tkt to log-in or log-out
        to a Drupal user.'''
        is_ckan_cookie, drupal_session_id = self._parse_cookies(environ)

        log.debug("Is cookie "+str(is_ckan_cookie)+" Dru ses ID:"+str(drupal_session_id))
        # Is there a Drupal cookie? We may want to do a log-in for it.
        if drupal_session_id:
            # Look at any authtkt logged in user details
            authtkt_identity = environ.get('repoze.who.identity')
            if authtkt_identity:
                authtkt_user_name = authtkt_identity['repoze.who.userid'] #same as environ.get('REMOTE_USER', '')
                authtkt_drupal_session_id = authtkt_identity['userdata']
            else:
                authtkt_user_name = ''
                authtkt_drupal_session_id = ''

            if not authtkt_user_name:
                # authtkt not logged in, so log-in with the Drupal cookie
                self._do_drupal_login(environ, drupal_session_id, new_headers)
                return
            elif authtkt_drupal_session_id.startswith(self._drupal_session_prefix):
                # A drupal user is logged in with authtkt.
                authtkt_drupal_session_id = authtkt_drupal_session_id[len(self._drupal_session_prefix):]
                # See if that the authtkt matches the drupal cookie's session
                if authtkt_drupal_session_id != drupal_session_id:
                    # Drupal cookie session has changed, so tell authkit to forget the old one
                    # before we do the new login
                    log.debug('Drupal cookie session has changed.')
                    #log.debug('Drupal cookie session has changed from %r to %r.', authtkt_drupal_session_id, drupal_session_id)
                    self._log_out(environ, new_headers)
                    self._do_drupal_login(environ, drupal_session_id, new_headers)
                    #log.debug('Headers on log-out log-in result: %r', new_headers)
                    return
                else:
                    log.debug('Drupal cookie session stayed the same.')
                    # Drupal cookie session matches the authtkt - leave user logged ini

                    # Just check that authtkt cookie is not too old - in the
                    # mean-time, Drupal may have invalidated the user, for example.
                    if self.is_authtkt_cookie_too_old(authtkt_identity):
                        log.info('Rechecking Drupal cookie')
                        self._log_out(environ, new_headers)
                        self._do_drupal_login(environ, drupal_session_id, new_headers)
                    return
            else:
                # There's a Drupal cookie, but user is logged in as a normal CKAN user.
                # Ignore the Drupal cookie.
                return
        elif not drupal_session_id and is_ckan_cookie:
            # Deal with the case where user is logged out of Drupal
            # i.e. user WAS were logged in with Drupal and the cookie was
            # deleted (probably because Drupal logged out)
            
            # Is the logged in user a Drupal user?
            user_name = environ.get('REMOTE_USER', '')
            if user_name:
                log.debug('Was logged in as Drupal user %r but Drupal cookie no longer there.', user_name)
                self._log_out(environ, new_headers)


    def _do_drupal_login(self, environ, drupal_session_id, new_headers):
        '''Given a Drupal cookie\'s session ID, check it with Drupal, create/modify
        the equivalent CKAN user with properties copied from Drupal and log the
        person in with auth_tkt and its cookie.
        '''
        if not hasattr(self.localthread, 'drupal_client') or self.localthread.drupal_client is None:
            self.localthread.drupal_client = DrupalClient()
        # ask drupal for the drupal_user_id for this session
        try:
            drupal_user_id = self.localthread.drupal_client.get_user_id_from_session_id(drupal_session_id)
        except DrupalRequestError, e:
            log.error('Error checking session with Drupal: %s', e)
            return
        if not drupal_user_id:
            log.debug('Drupal said the session ID found in the cookie is not valid.')
            return

        # ask drupal about this user
        drupal_user_properties = self.localthread.drupal_client.get_user_properties(drupal_user_id)
        user_dict = DrupalUserMapping.drupal_user_to_ckan_user(drupal_user_properties)

        # see if user already exists in CKAN
        ckan_user_name = user_dict['name']
        from ckan import model
        from ckan.model.meta import Session
        from ckanext.dgu.lib.user_helpers import merge_user
        query = Session.query(model.User).filter_by(id=drupal_user_id)
        if not query.count():
            query = Session.query(model.User).filter_by(name=ckan_user_name)
        if not query.count():
            # need to add this user to CKAN
            user = model.User(**user_dict)
            Session.add(user)
            Session.commit()
            log.debug('Drupal user added to CKAN as: %s', user.name)
            merge_user(user, user.name)
        else:
            user = query.one()
            log.debug('Drupal user found in CKAN: %s', user.name)
            
            current_name = user.name

            if user.id != drupal_user_id or \
                    user.email != user_dict['email'] or \
                    user.name != user_dict['name'] or \
                    user.fullname != user_dict['fullname'] or \
                    user.state == 'deleted':
                user.id = drupal_user_id
                user.email = user_dict['email']
                user.name = user_dict['name']
                user.fullname = user_dict['fullname']
                if user.state == 'deleted':
                    log.debug('User was deleted, changing state back to active.')
                    user.state = 'active'
                log.debug('User details updated from Drupal: %s %s %s %s',
                          user.id, user.email, user.name, user.fullname)
                model.Session.commit()
                
            if current_name != user_dict['name']: 
                #We encountered a problem where users were duplicated in Drupal, using same
                #emails etc. but different usernames due to using different certificates to login
                #Since the fix, a change of username is a trigger for possible merges of these
                #duplicated users as well, because we cannot do this safely with a general DB
                #update and we instead merge inside CKAN on first login with new format of username
                
                #merge data from all users with same email + a couple of exceptions where their
                #emails actually differ but we know they are the same users
                merge_user(user, current_name)
                

        self.set_roles(ckan_user_name, drupal_user_properties['roles'].values())
        log.debug("Keys: "+str(drupal_user_properties.keys()))
        if 'orgs' in drupal_user_properties.keys():
            self.set_organizations(ckan_user_name, drupal_user_properties['orgs'].values())

        # There is a chance that on this request we needed to get authtkt
        # to log-out. This would have created headers like this:
        #   'Set-Cookie', 'auth_tkt="INVALID"...'
        # but since we are about to login again, which will create a header
        # setting that same cookie, we need to get rid of the invalidation
        # header first.
        new_headers[:] = [(key, value) for (key, value) in new_headers \
                            if (not (key=='Set-Cookie' and value.startswith('auth_tkt="INVALID"')))]
        #log.debug('Headers reduced to: %r', new_headers)

        # Ask auth_tkt to remember this user so that subsequent requests
        # will be authenticated by auth_tkt.
        # auth_tkt cookie template needs to also go in the response.
        identity = {'repoze.who.userid': str(ckan_user_name),
                    'tokens': '',
                    'userdata': self._drupal_session_prefix+drupal_session_id}
        headers = environ['repoze.who.plugins']['dgu_auth_tkt'].remember(environ, identity)
        if headers:
            new_headers.extend(headers)

        # Tell app during this request that the user is logged in
        environ['REMOTE_USER'] = user.name
        log.debug('Set REMOTE_USER = %r', user.name)

    def set_roles(self, user_name, drupal_roles):
        '''Sets CKAN user roles based on the drupal roles.

        Restricted to sysadmin. Publishing roles initially imported during migration from
        Drupal.

        Example drupal_roles:
        ['package admin', 'publisher admin', 'authenticated user', 'publishing user']
        where sysadmin roles are:
               3   'administrator' - total control
               11  'package admin' - admin of datasets
                   'publisher admin' - admin of publishers
        other roles:
                   'publishing user' - anyone who has registered - includes spammers
        '''
        from ckan import model
        from ckan import new_authz
        import ckan.plugins.toolkit as tk
        needs_commit = False
        user = model.User.by_name(user_name)

        context = {'user':'frontend'}
        ckan_roles = []
        
        #from pylons import config
        #from pylons.i18n import _
        #from pylons.i18n.translation import set_lang
        #set_lang('en_GB')
        
        for role in self.SYNCED_ROLES:
            if role['drupal_name'] in drupal_roles:
                try:
                    result = tk.get_action('group_member_create')(
                    context=context, data_dict={'id': role['id'], 'username': user_name, 'role': 'member'})
                except (TypeError):
                    self._init_translator()
                    result = tk.get_action('group_member_create')(
                    context=context, data_dict={'id': role['id'], 'username': user_name, 'role': 'member'})
                ckan_roles.append(role['title'])
            else:
                try:
                    result = tk.get_action('group_member_delete')(
                    context=context, data_dict={'id': role['id'], 'username': user_name})
                except (TypeError):
                    self._init_translator()
                    result = tk.get_action('group_member_delete')(
                    context=context, data_dict={'id': role['id'], 'username': user_name})
        log.debug('Drupal roles %r, as groups in CKAN: %r', drupal_roles, ckan_roles)
        
        # Sysadmin or not
        should_be_sysadmin = bool(set(('administrator', 'package admin', 'publisher admin', 'ckan administrator', 'CKAN sysadmin', 'Sistemski administrator')) & set(drupal_roles))
        if should_be_sysadmin and not user.sysadmin:
            # Make user a sysadmin
            user.sysadmin = True
            log.info('User made a sysadmin: %s', user_name)
            needs_commit = True
        elif not should_be_sysadmin and user.sysadmin:
            # Stop user being a sysadmin - disabled for time being which 'ckan
            # administrator' is populated
            user.sysadmin = False
            log.info('User now not a sysadmin: %s', user_name)
            needs_commit = True
            pass
        if needs_commit:
            model.repo.commit_and_remove()
    
    def set_organizations(self, user_name, drupal_orgs):
        '''Sets CKAN user roles based on the drupal roles.

        Restricted to sysadmin. Publishing roles initially imported during migration from
        Drupal.

        Example drupal_roles:
        ['package admin', 'publisher admin', 'authenticated user', 'publishing user']
        where sysadmin roles are:
               3   'administrator' - total control
               11  'package admin' - admin of datasets
                   'publisher admin' - admin of publishers
        other roles:
                   'publishing user' - anyone who has registered - includes spammers
        '''
        from ckan import model
        from ckan import new_authz
        import ckan.plugins.toolkit as tk
        needs_commit = False
        user = model.User.by_name(user_name)

        log.debug('User orgs in Drupal: %r', drupal_orgs)
        context = {'user':'frontend'}
        ckan_orgs = []

    def is_authtkt_cookie_too_old(self, authtkt_identity):
        authtkt_time = datetime.datetime.fromtimestamp(authtkt_identity['timestamp'])
        age = datetime.datetime.now() - authtkt_time
        age_in_seconds = age.seconds + age.days * 24 * 3600
        log.debug('Seconds since checking Drupal cookie: %s (threshold=%s)', age_in_seconds, self.seconds_between_checking_drupal_cookie)
        return age_in_seconds > self.seconds_between_checking_drupal_cookie

class DrupalUserMapping:
    _user_name_prefix = 'user_d'

    @classmethod
    def drupal_user_to_full_name(cls, drupal_user_dict):
        from HTMLParser import HTMLParser
        html_parser = HTMLParser()
        if not drupal_user_dict:
            return None
         
        try:
            firstname = drupal_user_dict['field_first_name']['und'][0]['safe_value']
            firstname = html_parser.unescape(firstname)
        except:
            firstname = ''
        
        try:
            surname = drupal_user_dict['field_surname']['und'][0]['safe_value']
            surname = html_parser.unescape(surname)
        except:
            surname = ''
        
        if firstname and surname:
            return firstname + " " + surname
        elif firstname:
            return firstname
        elif surname:
            return surname
        else:
            return drupal_user_dict['name']
    
    @classmethod
    def drupal_id_to_ckan_user_name(cls, drupal_id):
        # Drupal ID is always a number
        drupal_id.lower().replace(' ', '_') # just in case
        return u'%s%s' % (cls._user_name_prefix, drupal_id)

    @classmethod
    def ckan_user_name_to_drupal_id(cls, ckan_user_name):
        if ckan_user_name.startswith(cls._user_name_prefix):
            return ckan_user_name[len(cls._user_name_prefix):]
        else:
            return None # Not a Drupal user

    @classmethod
    def drupal_user_to_ckan_user(cls, drupal_user_dict, existing_user_name=None):
        return dict(
            id=drupal_user_dict['uid'],
            name=existing_user_name or
                drupal_user_dict['name'],
            fullname=unicode(DrupalUserMapping.drupal_user_to_full_name(drupal_user_dict)),
            about=u'User account imported from Drupal system.',
            email=drupal_user_dict['mail'],
            created=datetime.datetime.fromtimestamp(int(drupal_user_dict['created']))
            )

