﻿# -*- coding: utf-8 -*-

import re
import json
import logging
from itertools import chain

from ckan.lib.base import c, model
from ckan.lib.field_types import DateType, DateConvertError
from ckan.lib.navl.dictization_functions import Invalid

import ckan.logic.schema as default_schema
import ckan.logic.validators as val

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckanext.dgu.lib import publisher as publib
from ckanext.dgu.lib import helpers as dgu_helpers
from ckanext.dgu.forms.validators import merge_resources, unmerge_resources, \
     validate_resources, \
     validate_additional_resource_types, \
     validate_data_resource_types, \
     validate_license, \
     drop_if_same_as_publisher, \
     populate_from_publisher_key_if_missing, \
     populate_as_publisher_title_if_missing, \
     populate_from_publisher_if_missing, \
     populate_from_user_key_if_missing, \
     remove_blank_resources, \
     to_json, from_json, \
     bool_, \
     value_if_missing, \
     opsi_owner_org_validator, \
     not_empty, \
     not_empty_if_timeseries, \
     tag_string_convert, \
     add_to_tag_string, \
     to_lowercase, \
     ignore_not_allowed
from ckan.lib.navl.dictization_functions import missing

from ckanext.dgu.forms.validators import tag_string_convert as opsi_tag_string_convert
from pylons import config

import logging
log = logging.getLogger(__name__)

#convert_from_extras = tk.get_validator('convert_from_extras')
ignore_missing = tk.get_validator('ignore_missing')
#not_empty = tk.get_validator('not_empty')
#not_empty = ckanext.dgu.forms.validators.not_empty
empty = tk.get_validator('empty')
ignore = tk.get_validator('ignore')
ignore_empty = tk.get_validator('ignore_empty')
keep_extras = tk.get_validator('keep_extras')
not_missing = tk.get_validator('not_missing')
free_tags_only = tk.get_converter('free_tags_only')
convert_to_tags = tk.get_converter('convert_to_tags')
convert_from_tags = tk.get_converter('convert_from_tags')

geographic_granularity = [('', ''),
                          (u'nacionalna', u'nacionalna'),
                          (u'regionalna', u'regionalna'),
                          (u'lokalna avtoriteta', u'lokalna avtoriteta'),
                          (u'oddelek', u'oddelek'),
                          (u'točka', u'točka'),
                          (u'ostalo - specificirajte', u'ostalo - specificirajte')]

update_frequency = [(u'stalno', u'stalno'),
                    (u'dnevno', u'dnevno'),
                    (u'tedensko', u'tedensko'),
                    (u'dvotedensko', u'dvotedensko'),
                    (u'mesečno', u'mesečno'),
                    (u'četrtletno', u'četrtletno'),
                    (u'polletno', u'polletno'),
                    (u'letno', u'letno'),
                    (u'po potrebi', u'po potrebi'),
                    (u'neredno', u'neredno'),
                    (u'ni načrtovano', u'ni načrtovano')]

temporal_granularity = [("", ""),
                        (u'letno', u'letno'),
                        (u'četrtletno', u'četrtletno'),
                        (u'mesečno', u'mesečno'),
                        (u'tedensko', u'tedensko'),
                        (u'dnevno', u'dnevno'),
                        (u'vsako uro', u'vsako uro'),
                        (u'točka', u'točka'),
                        (u'ostalo - specificirajte', u'ostalo - specificirajte')]


def resources_schema():
    schema = default_schema.default_resource_schema()
    # don't convert values to ints - that doesn't work for extensions
    # (see conversation with John Glover)
    schema.update({
        'created': [ignore_missing],
        'position': [ignore_missing],
        'last_modified': [ignore_missing],
        'cache_last_updated': [ignore_missing],
        'webstore_last_updated': [ignore_missing],
        'datastore_active': [ignore_missing]
    })
    return schema


def resources_schema_to_form():
    schema = resources_schema()
    schema['date'] = [ignore_missing, date_to_form]
    return schema


def resources_schema_to_db():
    schema = resources_schema()
    schema['id'].append(new_resource_if_url_and_description_change)
    schema['format'] = [not_empty, unicode]
    schema['url'] = [not_empty, limit_url_changes_if_uploaded]
    schema['description'] = [not_empty]
    return schema


def additional_resource_schema_to_db():
    schema = resources_schema_to_db()
    schema['resource_type'].insert(0, validate_additional_resource_types)
    return schema


def individual_resource_schema_to_db():
    schema = resources_schema_to_db()
    schema['resource_type'].insert(0, validate_data_resource_types)
    return schema


def timeseries_resource_schema_to_db():
    schema = resources_schema_to_db()
    schema['date'] = [not_empty, unicode, convert_to_extras, date_to_db]
    schema['resource_type'].insert(0, validate_data_resource_types)
    return schema


class DatasetForm(p.SingletonPlugin):

    p.implements(p.IDatasetForm, inherit=True)

    # We don't customize the schema here - instead it is done in the validate
    # function, because there it has the context.
    #def create_package_schema(self):
    #def update_package_schema(self):

    def show_package_schema(self):
        return self.db_to_form_schema()

    def new_template(self):
        return 'package/new.html'

    def edit_template(self):
        return 'package/edit.html'

    def comments_template(self):
        return 'package/comments.html'

    def search_template(self):
        return 'package/search.html'

    def read_template(self):
        return 'package/read.html'

    def history_template(self):
        return 'package/history.html'

    def is_fallback(self):
        return True

    def package_types(self):
        return ["dataset"]

    def package_form(self, package_type=None):
        return 'package/edit_form.html'

    def setup_template_variables(self, context, data_dict=None,
                                 package_type=None):
        is_edit_or_new_form = 'save' in context
        if is_edit_or_new_form:
            # These expensive calls are needed for the edit/new form (not
            # package read).
            # It's not ideal to use c - normally use a helper - but it makes
            # sense here since several different templates use c.publishers.
            c.publishers = self.get_publishers()
            c.publishers_json = json.dumps(c.publishers)

    # Override the form validation to be able to vary the schema by the type of
    # package and user
    def validate(self, context, data_dict, schema, action):
        #log.debug("validate before if, action: " + action)
        if action in ('package_update', 'package_create'):
            # If the caller to package_update specified a schema (e.g.
            # harvesters specify the default schema) then we don't want to
            # override that.
            #log.debug("validating " + str(data_dict) + " context: " + str(context))
            if not context.get('schema'):
                schema = self.form_to_db_schema_options(context)
                if 'api_version' in context:
                    # Tag validation is looser than CKAN default
                    schema['tags'] = tags_schema()
        return tk.navl_validate(data_dict, schema, context)

    def form_to_db_schema_options(self, context):
        '''Returns the schema for the customized DGU form.'''
        schema = self.form_to_db_schema()
        # UKLP datasets have looser validation constraints.  This is because
        # UKLP datasets are created using a custom schema passed in from the
        # harvester.  However, when it comes to re-saving the dataset via the
        # dataset form or API, there are some validation requirements we need
        # to drop.  That's what this section of code does.
        pkg = context.get('package')
        if pkg and pkg.extras.get('UKLP') == 'True':
            self._uklp_schema_updates(schema)
        if pkg and pkg.extras.get('external_reference') == 'ONSHUB':
            self._ons_schema_updates(schema)
        return schema

    def _uklp_schema_updates(self, schema):
        schema.update(
          {
            'notes': [ignore_missing, unicode],
            'theme-primary': [ignore_missing, unicode, convert_to_extras],
            'temporal_coverage-from': [ignore_missing, unicode, convert_to_extras],
            'temporal_coverage-to': [ignore_missing, unicode, convert_to_extras],
            'groups': {
                'name': [ignore_missing, validate_group_id_or_name_exists_if_not_blank, unicode],
                'id': [ignore_missing, unicode],
            },
          }
          )
        for resources in ('additional_resources',
                          'timeseries_resources',
                          'individual_resources'):
            schema[resources]['format'] = [unicode]  # i.e. optional
        # in addition to these, validate_license() allows no licence

    def _ons_schema_updates(self, schema):
        schema.update(
            {
                'theme-primary': [ignore_missing, unicode, convert_to_extras],
            })
        for resources in ('additional_resources',
                          'timeseries_resources',
                          'individual_resources'):
            schema[resources]['format'] = [unicode]  # i.e. optional

    def db_to_form_schema_options(self, options={}):
        context = options.get('context', {})
        schema = context.get('schema', None)
        if schema:
            return schema
        else:
            return self.db_to_form_schema()

    @classmethod
    def form_to_db_schema(cls, package_type=None):
        schema = {
            'title': [not_empty, unicode],
            'name': [not_empty, unicode, val.name_validator, val.package_name_validator],
            'notes': [not_empty, unicode],

            'date_released': [ignore_missing, date_to_db, convert_to_extras],
            'date_updated': [ignore_missing, date_to_db, convert_to_extras],
            'date_update_future': [ignore_missing, date_to_db, convert_to_extras],
            'last_major_modification': [ignore_missing, date_to_db, convert_to_extras],
            'precision': [ignore_missing, unicode, convert_to_extras],
            'geographic_granularity': [ignore_missing, use_other, unicode, convert_to_extras],
            'geographic_granularity-other': [ignore_missing],
            #'geographic_coverage': [ignore_missing, convert_geographic_to_db, convert_to_extras],
            'geographic_coverage': [not_empty, unicode, convert_to_extras],
            'pogostost_osvezevanja': [not_empty_if_timeseries, ignore_missing, unicode, convert_to_extras],
            'temporal_granularity': [ignore_missing, use_other, unicode, convert_to_extras],
            'temporal_granularity-other': [ignore_missing],
            'temporal_coverage-from': [ignore_missing, date_to_db, convert_to_extras],
            'temporal_coverage-to': [ignore_missing, date_to_db, convert_to_extras],
            'url': [ignore_missing, unicode],
            'taxonomy_url': [ignore_missing, unicode, convert_to_extras],

            'additional_resources': additional_resource_schema_to_db(),
            'timeseries_resources': timeseries_resource_schema_to_db(),
            'individual_resources': individual_resource_schema_to_db(),

            'owner_org': [not_empty, opsi_owner_org_validator, unicode],
            'groups': {
                'name': [not_empty, val.group_id_or_name_exists, unicode],
                'id': [ignore_missing, unicode],
            },

            'opsi_tip': [ignore_missing, unicode, convert_to_extras],
            'opsi_tip_naziv': [ignore_missing, unicode, convert_to_extras],
            'contact-name': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            'contact-email': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            'contact-phone': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            
            #Odgovorna oseba
            'foi-name': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            'foi-email': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            'foi-phone': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],
            'foi-web': [ignore_missing, unicode, drop_if_same_as_publisher, convert_to_extras],

            'notification_recipient': [ignore_missing, unicode, convert_to_extras],

            'published_via': [ignore_missing, unicode, convert_to_extras],

            'licence': [not_empty, unicode, convert_to_extras],
            #'licence_in_form': [ignore_missing, unicode], 

            #'licenca': [ignore_missing, convert_to_extras],
            #'licenca_in_form': [ignore_missing, convert_to_extras],

            'tags': tags_schema(),
            'tag_string': [ignore_missing, opsi_tag_string_convert],
            
            #'kljucne_besede': [ignore_missing, commas_to_list, to_json, convert_to_extras],
            
            'creator_user_id': [ignore_missing, ignore_not_allowed],
            'maintainer_email': [ignore_missing, ignore_not_allowed],
            'state': [value_if_missing('draft'), ignore_not_allowed],

            'unpublished': [ignore_missing, bool_, convert_to_extras],
            
            'core-dataset': [ignore_missing, bool_, convert_to_extras],
            'release-notes': [ignore_missing, unicode, convert_to_extras],
            'publish-date': [ignore_missing, date_to_db, convert_to_extras],
            'publish-restricted': [ignore_missing, bool_, convert_to_extras],
            
            'podrocje': [not_empty, unicode, convert_to_extras],
            
            #'dostopnost_id': [unicode, convert_to_extras],
            #'dostopnost_desc': [unicode,  convert_to_extras],
            'dostopnost_zbirke': [not_empty, unicode,  convert_to_extras],
            
            'jezik_zbirke': [not_empty, unicode,  convert_to_extras],
            
            'povezanost_desc': [value_if_missing('Ni povezave'), unicode, convert_to_extras],
            
            'vrsta_pravne_podlage': [ignore_missing, unicode,  convert_to_extras],
            'sop_stevilka': [ignore_missing, unicode, convert_to_extras],
            'naziv_zakona': [ignore_missing, unicode, convert_to_extras],
            #'naziv_zakona_container': [unicode, convert_to_extras],
            'naziv_zakona_lep_izpis': [ignore_missing, unicode, convert_to_extras],
            
            'pogoji_licence': [ignore_missing, unicode,  convert_to_extras],
            'pogoji_licence_cena': [ignore_missing, unicode,  convert_to_extras],
            
            'municipal': [value_if_missing(False), bool_, convert_to_extras],
            'high_value_data': [value_if_missing(False), bool_, convert_to_extras],
            'bulk_download': [value_if_missing(False), bool_, convert_to_extras],

            'theme-primary': [ignore_missing, unicode, convert_to_extras],
            'theme-secondary': [ignore_missing, commas_to_list, to_json, convert_to_extras],
            'extras': default_schema.default_extras_schema(),

            # This is needed by the core CKAN update_resource, but isn't found by it because
            # we do the work in __after.
            'resources': resources_schema(),

            '__extras': [ignore],
            '__junk': [empty],
            '__after': [validate_license, remove_blank_resources, validate_resources, merge_resources]
        }
        return schema

    def db_to_form_schema(data, package_type=None):
        schema = {
            'date_released': [convert_from_extras, ignore_missing, date_to_form],
            'date_updated': [convert_from_extras, ignore_missing, date_to_form],
            'last_major_modification': [convert_from_extras, ignore_missing, date_to_form],
            'date_update_future': [convert_from_extras, ignore_missing, date_to_form],
            'precision': [convert_from_extras, ignore_missing],
            'geographic_granularity': [convert_from_extras, ignore_missing, extract_other(geographic_granularity)],
            #'geographic_coverage': [convert_from_extras, ignore_missing, convert_geographic_to_form],
            'geographic_coverage': [convert_from_extras, ignore_missing, unicode],
            'pogostost_osvezevanja': [convert_from_extras, ignore_missing, unicode],
            'temporal_granularity': [convert_from_extras, ignore_missing, extract_other(temporal_granularity)],
            'temporal_coverage-from': [convert_from_extras, ignore_missing, date_to_form],
            'temporal_coverage-to': [convert_from_extras, ignore_missing, date_to_form],
            'taxonomy_url': [convert_from_extras, ignore_missing],

            'resources': resources_schema_to_form(),
            'extras': {
                'key': [],
                'value': [],
                '__extras': [keep_extras]
            },
            'tags': {
                '__extras': [keep_extras, free_tags_only, add_to_tag_string]
            },
            #filled in package_show:
            'tag_string': [],
                  
            'licence': [convert_from_extras, ignore_missing],
                  
            #'licenca': [convert_from_extras, ignore_missing],
            #'licenca_in_form': [convert_from_extras, ignore_missing],
                  
            #'kljucne_besede': [ignore_missing, from_json, list_to_comma, convert_from_extras],

            'organization': [],
            'owner_org': [],
            'creator_user_id': [ignore_missing],
            'maintainer_email': [ignore_missing],

            'groups': {
                'name': [ignore_missing, unicode]
            },

            'opsi_tip': [convert_from_extras, populate_from_publisher_key_if_missing('opsi_tip', False), ignore_missing],
            'opsi_tip_naziv': [convert_from_extras, populate_from_publisher_key_if_missing('opsi_tip_naziv', False), ignore_missing],
            'contact-name': [convert_from_extras, populate_as_publisher_title_if_missing, ignore_missing],
            'contact-email': [convert_from_extras, populate_from_publisher_key_if_missing('el_posta'), ignore_missing],
            'contact-phone': [convert_from_extras, populate_from_publisher_key_if_missing('telefon'), ignore_missing],

            'foi-name': [convert_from_extras, populate_from_user_key_if_missing('fullname'), ignore_missing],
            'foi-email': [convert_from_extras, populate_from_user_key_if_missing('email'), populate_from_publisher_key_if_missing('el_posta'), ignore_missing],
            'foi-phone': [convert_from_extras, populate_from_publisher_key_if_missing('telefon'), ignore_missing],
            'foi-web': [convert_from_extras, populate_from_publisher_key_if_missing('url_naslov'), ignore_missing],

            'notification_recipient': [convert_from_extras, ignore_missing],

            'unpublished': [convert_from_extras, ignore_missing],
            
            'core-dataset': [convert_from_extras, ignore_missing],
            'release-notes': [convert_from_extras, ignore_missing],
            'publish-date': [convert_from_extras, ignore_missing, date_to_form],
            'publish-restricted': [convert_from_extras, ignore_missing],
            
            'podrocje': [convert_from_extras, ignore_missing],
            
            #'dostopnost_id': [convert_from_extras],
            #'dostopnost_desc': [convert_from_extras, ignore_missing],
            'dostopnost_zbirke': [convert_from_extras, ignore_missing],
            
            'jezik_zbirke': [convert_from_extras, ignore_missing],
            
            'povezanost_desc': [convert_from_extras, ignore_missing],
            
            'vrsta_pravne_podlage': [convert_from_extras, ignore_missing],
            'sop_stevilka': [convert_from_extras, ignore_missing],
            'naziv_zakona': [convert_from_extras, ignore_missing],
            #'naziv_zakona_container': [convert_from_extras, ignore_missing],
            'naziv_zakona_lep_izpis': [convert_from_extras, ignore_missing],
            
            'pogoji_licence': [convert_from_extras, ignore_missing],
            'pogoji_licence_cena': [convert_from_extras, ignore_missing],
            
            'municipal': [convert_from_extras, value_if_missing(False)],
            'high_value_data': [convert_from_extras, value_if_missing(False)],
            'bulk_download': [convert_from_extras, value_if_missing(False)],
            'theme-primary': [convert_from_extras, ignore_missing],
            'theme-secondary': [convert_from_extras, from_json, ignore_missing],
            '__after': [unmerge_resources],
            '__extras': [keep_extras],
            '__junk': [ignore],
        }
        return schema

    def check_data_dict(self, data_dict, package_type=None):
        return

    def get_publishers(self):
        from ckan.model.group import Group

        log.debug('Getting publishers')

        if dgu_helpers.is_sysadmin() or dgu_helpers.is_executive_editor():
            groups = Group.all(group_type='organization')
            log.debug("Retrieved sysadmin publishers.")
        elif c.userobj:
            # need to get c.userobj again as it may be detached from the
            # session since the last time we called get_groups (it caches)
            c.userobj = model.User.by_name(c.user)

            # For each group where the user is an admin, we should also include
            # all of the child publishers.
            log.debug("Getting user groups")
            admin_groups = set()
            for g in c.userobj.get_groups('organization', 'admin'):
                for pub in publib.go_down_tree(g):
                    admin_groups.add(pub)
            log.debug("Got user groups")

            editor_groups = c.userobj.get_groups('organization', 'editor')
            groups = list(admin_groups) + editor_groups
        else:  # anonymous user shouldn't have access to this page anyway.
            groups = []

        log.debug("Generating group list")
        # Be explicit about which fields we make available in the template
        group_dict = dict((g.id, {
            'name': g.name,
            'id': g.id,
            'title': g.title
        }) for g in groups)
        
        log.debug("Retrieved publishers, returning them.")

        return group_dict


def date_to_db(value, context):
    try:
        value = DateType.form_to_db(value)
    except DateConvertError, e:
        raise Invalid(str(e))
    return value


def date_to_form(value, context):
    try:
        value = DateType.db_to_form(value)
    except DateConvertError, e:
        raise Invalid(str(e))
    return value


def convert_to_extras(key, data, errors, context):

    current_index = max([int(k[1]) for k in data.keys()
                         if len(k) == 3 and k[0] == 'extras'] + [-1])

    data[('extras', current_index+1, 'key')] = key[-1]
    data[('extras', current_index+1, 'value')] = data[key]


def convert_from_extras(key, data, errors, context):
    for data_key, data_value in data.iteritems():
        if (data_key[0] == 'extras'
            and data_key[-1] == 'key'
            and data_value == key[-1]):
            data[key] = data[('extras', data_key[1], 'value')]

def remove_blanks(key, data, errors, context):
    unfiltered = data[key]
    if isinstance(unfiltered, basestring):
        return

    try:
        filtered = [x for x in unfiltered if x != ""]
        data[key] = filtered
    except:
        pass

def to_list(key, data, errors, context):
    item = data[key]
    if isinstance(item, basestring):
        data[key] = [item]

def commas_to_list(key, data, errors, context):
    item = data[key]
    if isinstance(item, basestring):
        data[key] = [i.strip() for i in item.split(',')]
        
def list_to_comma(key, data, errors, context):
    item = eval(data[key])
    if isinstance(item, basestring):
        data[key] = ','.join(item)


def use_other(key, data, errors, context):

    other_key = key[-1] + '-other'
    other_value = data.get((other_key,), '')
    if isinstance(other_value, basestring):
        other_value = other_value.strip()
    if other_value:
        data[key] = other_value


def extract_other(option_list):

    def other(key, data, errors, context):
        value = data[key]
        if value in dict(option_list).keys():
            return
        elif value is missing:
            data[key] = ''
            return
        else:
            data[key] = 'other'
            other_key = key[-1] + '-other'
            data[(other_key,)] = value
    return other


def convert_geographic_to_db(value, context):
    from ckanext.dgu.schema import GeoCoverageType

    if isinstance(value, list):
        regions = value
    elif value:
        regions = [value]
    else:
        regions = []

    return GeoCoverageType.get_instance().form_to_db(regions)


def convert_geographic_to_form(value, context):
    from ckanext.dgu.schema import GeoCoverageType
    return GeoCoverageType.get_instance().db_to_form(value)


def validate_group_id_or_name_exists_if_not_blank(value, context):
    if not value.strip():
        return True
    return val.group_id_or_name_exists(value, context)


def tag_name_extended_validator(value, context):
    tagname_match = re.compile('[\w \-_.()":&/,+=\[\]*]*$', re.UNICODE)
    if not tagname_match.match(value):
        raise Invalid(('Tag "%s" must be alphanumeric '
            'characters or symbols: -_.()":&/,+=[]*') % (value))
    return value

def tags_schema():
    # Allow tags of 1 character e.g. B (chemical name for Boron)
    schema = {
        'name': [not_missing,
                 not_empty,
                 unicode,
                 tag_name_extended_validator,
                 ],
        'revision_timestamp': [ignore],
        'state': [ignore],
    }
    return schema

def id_to_dict(key, data, errors, context):
    from ckanext.dgu.model.schema_codelist import Schema, Codelist
    for i, id_ in enumerate(data[key]):
        if key == ('schema',):
            obj = Schema.get(id_)
        elif key == ('codelist',):
            obj = Codelist.get(id_)
        else:
            raise NotImplementedError('Bad key: %s' % key)
        if not obj:
            raise Invalid('%s id does not exist: %s' % (key, id_))
        data[key][i] = obj.as_dict()

def schema_codelist_validator(key, data, errors, context):
    from ckanext.dgu.model.schema_codelist import Schema, Codelist
    for i, schema_ref in enumerate(data[key]):
        if not schema_ref:
            # drop-down has no selection - ignore
            continue
        # form gives an ID. API might give a title.
        if key == ('schema',):
            obj = Schema.get(schema_ref) or Schema.by_title(schema_ref) or \
                    Schema.by_url(schema_ref)
        elif key == ('codelist',):
            obj = Codelist.get(schema_ref) or Codelist.by_title(schema_ref) or\
                    Codelist.by_url(schema_ref)
        else:
            raise NotImplementedError('Bad key: %s' % key)
        if not obj:
            raise Invalid('%s id does not exist: %r' % (key[0], schema_ref))
        # write the ID in case it came in via the API and was a URL or title
        data[key][i] = obj.id


def new_resource_if_url_and_description_change(key, data, errors, context):
    id_ = data[key]
    if not id_:
        return
    old_resource = model.Resource.get(id_)
    if not old_resource:
        return

    # Ignore checking for uploaded resources. Their URL should always be autogenerated,
    # but the filename can be changed freely (see validator limit_url_changes_if_uploaded).
    if old_resource.url_type == 'upload':
        return

    if data[(key[0], key[1], 'url')] != old_resource.url and \
            data[(key[0], key[1], 'description')] != old_resource.description:
        # Resource has changed so much we consider it a new one by deleting its
        # id and other hidden properties.  We saw occasions when a resource in
        # the form had its fields wiped and new description and URL put in, but
        # it kept the resource ID (hidden in the form) so it still had old
        # cached resources associated with it, incorrectly.
        resource_columns = set(('date', 'description', 'url', 'format'))
        for key_ in data.keys():
            if key_[0] == key[0] and \
                    key_[1] == key[1] and \
                    key_[2] not in resource_columns:
                del data[key_]

def limit_url_changes_if_uploaded(key, data, errors, context):
    from ckan.lib.base import h
    import ckan.lib.munge as munge
    from urlparse import urlparse

    url_ = data[key]
    if not url_:
        return
    id_ = data.get((key[0], key[1], 'id'))
    if not id_:
        return

    old_resource = model.Resource.get(id_)
    if not old_resource:
        return

    if old_resource.url_type == 'upload':
        resource_group = model.Session.query(model.ResourceGroup).get(old_resource.resource_group_id)
        last_part = url_.split('/')[-1]
        cleaned_name = munge.munge_filename(last_part)
        full_url = h.url_for(controller='package',
                        action='resource_download',
                        id=resource_group.package_id,
                        resource_id=id_,
                        filename=cleaned_name,
                        qualified=True)
        #url_for might still produce http or localhost domains instead of using site_url
        site_url = config.get('ckan.site_url', None)
        if site_url:
            data[key] = site_url + '/' + full_url.split('//',1)[-1].split('/',1)[-1]
        else:
            data[key] = full_url
