﻿# -*- coding: utf-8 -*-
import collections
import datetime
import logging
import os

from paste.deploy.converters import asbool
from pylons import config

from ckan import model
from ckan.lib.helpers import OrderedDict, url_for
from ckan.lib.celery_app import celery
import ckan.plugins as p
from ckanext.report import lib
from ckanext.dgu.lib.publisher import go_up_tree
from ckanext.dgu.lib import helpers as dgu_helpers

from sqlalchemy import or_

log = logging.getLogger(__name__)

def refresh_opsi_reports():
    from ckanext.report.report_registry import ReportRegistry
    import time
    registry = ReportRegistry.instance()
    timings = {}
    report_list = ['unpublished', 'draft']
    for report_name in report_list:
        s = time.time()
        registry.get_report(report_name).refresh_cache_for_all_options()
        timings[report_name] = time.time() - s
    log.debug("OPSI report refresh for reports %s complete %s", str(report_list), timings)

# asynchronously refreshes reports
def refresh_opsi_reports_async():
    log.debug("Asynchronously refreshing reports, sending task to celery.")
    ckan_ini_filepath = os.path.abspath(config['__file__'])
    celery.send_task("opsi.reports.refresh", args=[ckan_ini_filepath], queue="priority")

# NII
def nii_report():
    '''A list of the NII datasets, grouped by publisher, with details of broken
    links and source.'''
    nii_dataset_q = model.Session.query(model.Package)\
        .join(model.PackageExtra, model.PackageExtra.package_id == model.Package.id)\
        .join(model.Group, model.Package.owner_org == model.Group.id)\
        .filter(model.PackageExtra.key == 'core-dataset')\
        .filter(model.PackageExtra.value == 'true')\
        .filter(model.Package.state == 'active')
    nii_dataset_objects = nii_dataset_q\
            .order_by(model.Group.title, model.Package.title).all()

    def broken_resources_for_package(package_id):
        from ckanext.archiver.model import Archival

        results = model.Session.query(Archival, model.Resource)\
                       .filter(Archival.package_id == package_id)\
                       .filter(Archival.is_broken == True)\
                       .join(model.Package, Archival.package_id == model.Package.id)\
                       .filter(model.Package.state == 'active')\
                       .join(model.Resource, Archival.resource_id == model.Resource.id)\
                       .filter(model.Resource.state == 'active')

        broken_resources = [(resource.description, resource.id)
                            for archival, resource in results.all()]
        return broken_resources

    nii_dataset_details = []
    num_resources = 0
    num_broken_resources = 0
    num_broken_datasets = 0
    broken_organization_names = set()
    nii_organizations = set()
    for dataset_object in nii_dataset_objects:
        broken_resources = broken_resources_for_package(dataset_object.id)
        org = dataset_object.get_organization()
        dataset_details = {
                'name': dataset_object.name,
                'title': dataset_object.title,
                'dataset_notes': lib.dataset_notes(dataset_object),
                'organization_name': org.name,
                'unpublished': p.toolkit.asbool(dataset_object.extras.get('unpublished')),
                'num_broken_resources': len(broken_resources),
                'broken_resources': broken_resources,
                }
        nii_dataset_details.append(dataset_details)
        if broken_resources:
            num_broken_resources += len(broken_resources)
            num_broken_datasets += 1
            broken_organization_names.add(org.name)
        nii_organizations.add(org)
        num_resources += len(dataset_object.resources)

    org_tuples = [(org.name, org.title) for org in
                  sorted(nii_organizations, key=lambda o: o.title)]

    return {'table': nii_dataset_details,
            'organizations': org_tuples,
            'num_resources': num_resources,
            'num_datasets': len(nii_dataset_objects),
            'num_organizations': len(nii_organizations),
            'num_broken_resources': num_broken_resources,
            'num_broken_datasets': num_broken_datasets,
            'num_broken_organizations': len(broken_organization_names),
            }

nii_report_info = {
    'name': 'nii',
    'title': 'National Information Infrastructure',
    'description': 'Details of the datasets in the NII.',
    'option_defaults': OrderedDict([]),
    'option_combinations': None,
    'generate': nii_report,
    'template': 'report/nii.html',
}


# Publisher resources


def publisher_resources(organization=None,
                        include_sub_organizations=False):
    '''
    Returns a dictionary detailing resources for each dataset in the
    organisation specified.
    '''
    org = model.Group.by_name(organization)
    if not org:
        raise p.toolkit.ObjectNotFound('Publisher not found')

    # Get packages
    pkgs = model.Session.query(model.Package)\
                .filter_by(state='active')
    pkgs = lib.filter_by_organizations(pkgs, organization,
                                       include_sub_organizations).all()

    # Get their resources
    def create_row(pkg_, resource_dict):
        org_ = pkg_.get_organization()
        return OrderedDict((
                ('publisher_title', org_.title),
                ('publisher_name', org_.name),
                ('package_title', pkg_.title),
                ('package_name', pkg_.name),
                ('package_notes', lib.dataset_notes(pkg_)),
                ('resource_position', resource_dict.get('position')),
                ('resource_id', resource_dict.get('id')),
                ('resource_description', resource_dict.get('description')),
                ('resource_url', resource_dict.get('url')),
                ('resource_format', resource_dict.get('format')),
                ('resource_created', resource_dict.get('created')),
               ))
    num_resources = 0
    rows = []
    for pkg in pkgs:
        resources = pkg.resources
        if resources:
            for res in resources:
                res_dict = {'id': res.id, 'position': res.position,
                            'description': res.description, 'url': res.url,
                            'format': res.format,
                            'created': (res.created.isoformat()
                                        if res.created else None)}
                rows.append(create_row(pkg, res_dict))
            num_resources += len(resources)
        else:
            # packages with no resources are still listed
            rows.append(create_row(pkg, {}))

    return {'organization_name': org.name,
            'organization_title': org.title,
            'num_datasets': len(pkgs),
            'num_resources': num_resources,
            'table': rows,
            }

def publisher_resources_combinations():
    for organization in lib.all_organizations():
        for include_sub_organizations in (False, True):
            yield {'organization': organization}

publisher_resources_info = {
    'name': 'publisher-resources',
    'title': u'Datoteke organizacij',
    'description': u'Seznam zbirk in datotek po organizacijah.',
    'option_defaults': OrderedDict((('organization', 'ministrstvo_za_javno_upravo'),)),
    'option_combinations': publisher_resources_combinations,
    'generate': publisher_resources,
    'template': 'report/publisher_resources.html',
    }


def get_quarter_dates(datetime_now):
    '''Returns the dates for this (current) quarter and last quarter. Uses
    calendar year, so 1 Jan to 31 Mar etc.'''
    now = datetime_now
    month_this_q_started = (now.month - 1) // 3 * 3 + 1
    this_q_started = datetime.datetime(now.year, month_this_q_started, 1)
    this_q_ended = datetime.datetime(now.year, now.month, now.day)
    last_q_started = datetime.datetime(
                      this_q_started.year + (this_q_started.month-3)/12,
                      (this_q_started.month-4) % 12 + 1,
                      1)
    last_q_ended = this_q_started - datetime.timedelta(days=1)
    return {'this': (this_q_started, this_q_ended),
            'last': (last_q_started, last_q_ended)}


def publisher_activity(organization, include_sub_organizations=False):
    """
    Contains information about the datasets a specific organization has
    released in this and last quarter (calendar year). This is needed by
    departments for their quarterly transparency reports.
    """
    import datetime
    import ckan.model as model
    from paste.deploy.converters import asbool

    # These are the authors whose revisions we ignore, as they are trivial
    # changes. NB we do want to know about revisions by:
    # * harvest (harvested metadata)
    # * dgu (NS Stat Hub imports)
    # * Fix national indicators
    system_authors = ('autotheme', 'co-prod3.dh.bytemark.co.uk',
                      'Date format tidier', 'current_revision_fixer',
                      'current_revision_fixer2', 'fix_contact_details.py',
                      'Repoint 410 Gone to webarchive url',
                      'Fix duplicate resources',
                      'fix_secondary_theme.py',
                      )
    system_author_template = 'script-%'  # "%" is a wildcard

    created = {'this': [], 'last': []}
    modified = {'this': [], 'last': []}

    now = datetime.datetime.now()
    quarters = get_quarter_dates(now)

    if organization:
        organization = model.Group.by_name(organization)
        if not organization:
            raise p.toolkit.ObjectNotFound()

    if not organization:
        pkgs = model.Session.query(model.Package)\
                .all()
    else:
        pkgs = model.Session.query(model.Package)
        pkgs = lib.filter_by_organizations(pkgs, organization,
                                           include_sub_organizations).all()

    for pkg in pkgs:
        created_ = model.Session.query(model.PackageRevision)\
            .filter(model.PackageRevision.id == pkg.id) \
            .order_by("revision_timestamp asc").first()

        pr_q = model.Session.query(model.PackageRevision, model.Revision)\
            .filter(model.PackageRevision.id == pkg.id)\
            .filter_by(state='active')\
            .join(model.Revision)\
            .filter(~model.Revision.author.in_(system_authors)) \
            .filter(~model.Revision.author.like(system_author_template))
        rr_q = model.Session.query(model.Package, model.ResourceRevision, model.Revision)\
            .filter(model.Package.id == pkg.id)\
            .filter_by(state='active')\
            .join(model.ResourceGroup)\
            .join(model.ResourceRevision,
                  model.ResourceGroup.id == model.ResourceRevision.resource_group_id)\
            .join(model.Revision)\
            .filter(~model.Revision.author.in_(system_authors))\
            .filter(~model.Revision.author.like(system_author_template))
        pe_q = model.Session.query(model.Package, model.PackageExtraRevision, model.Revision)\
            .filter(model.Package.id == pkg.id)\
            .filter_by(state='active')\
            .join(model.PackageExtraRevision,
                  model.Package.id == model.PackageExtraRevision.package_id)\
            .join(model.Revision)\
            .filter(~model.Revision.author.in_(system_authors))\
            .filter(~model.Revision.author.like(system_author_template))

        for quarter_name in quarters:
            quarter = quarters[quarter_name]
            # created
            if quarter[0] < created_.revision_timestamp < quarter[1]:
                published = not asbool(pkg.extras.get('unpublished'))
                created[quarter_name].append(
                    (created_.name, created_.title, lib.dataset_notes(pkg),
                     'created', quarter_name,
                     created_.revision_timestamp.isoformat(),
                     created_.revision.author, published))

            # modified
            # exclude the creation revision
            period_start = max(quarter[0], created_.revision_timestamp)
            prs = pr_q.filter(model.PackageRevision.revision_timestamp > period_start)\
                        .filter(model.PackageRevision.revision_timestamp < quarter[1])
            rrs = rr_q.filter(model.ResourceRevision.revision_timestamp > period_start)\
                        .filter(model.ResourceRevision.revision_timestamp < quarter[1])
            pes = pe_q.filter(model.PackageExtraRevision.revision_timestamp > period_start)\
                        .filter(model.PackageExtraRevision.revision_timestamp < quarter[1])
            authors = ' '.join(set([r[1].author for r in prs] +
                                   [r[2].author for r in rrs] +
                                   [r[2].author for r in pes]))
            dates = set([r[1].timestamp.date() for r in prs] +
                        [r[2].timestamp.date() for r in rrs] +
                        [r[2].timestamp.date() for r in pes])
            dates_formatted = ' '.join([date.isoformat()
                                        for date in sorted(dates)])
            if authors:
                published = not asbool(pkg.extras.get('unpublished'))
                modified[quarter_name].append(
                    (pkg.name, pkg.title, lib.dataset_notes(pkg),
                        'modified', quarter_name,
                        dates_formatted, authors, published))

    datasets = []
    for quarter_name in quarters:
        datasets += sorted(created[quarter_name], key=lambda x: x[1])
        datasets += sorted(modified[quarter_name], key=lambda x: x[1])
    columns = ('Dataset name', 'Dataset title', 'Dataset notes', 'Modified or created', 'Quarter', 'Timestamp', 'Author', 'Published')

    quarters_iso = dict([(last_or_this, [date_.isoformat() for date_ in q_list])
                         for last_or_this, q_list in quarters.iteritems()])

    return {'table': datasets, 'columns': columns,
            'quarters': quarters_iso}

def publisher_activity_combinations():
    for org in lib.all_organizations(include_none=False):
        yield {'organization': org}
        #for include_sub_organizations in (False, True):
        #    yield {'organization': org,
        #           'include_sub_organizations': include_sub_organizations}

publisher_activity_report_info = {
    'name': 'publisher-activity',
    'title': u'Aktivnost organizacij',
    'description': u'Četrtletni popisi zbirk, ki jih je ustvarila določena organizacija.',
    'option_defaults': OrderedDict((('organization', 'ministrstvo_za_javno_upravo'),)),
#                                    ('include_sub_organizations', False),
#                                    )),
    'option_combinations': publisher_activity_combinations,
    'generate': publisher_activity,
    'template': 'report/publisher_activity.html',
    }


def unpublished():
    pkgs = model.Session.query(model.Package)\
                .filter(or_(model.Package.state=='draft', model.Package.state=='active'))\
                .join(model.PackageExtra)\
                .filter_by(key='unpublished')\
                .filter_by(value='true')\
                .all()
    pkg_dicts = []
    for pkg in pkgs:
        org = pkg.get_organization()
        podrocje = pkg.extras.get('podrocje', u'')
        author = dgu_helpers.get_user_name(pkg.creator_user_id)
        pkg_dict = {
                'id': pkg.id,
                'name': pkg.name,
                'title': pkg.title,
                'organization title': org.title,
                'organization name': org.name,
                'notes': pkg.notes,
                'author': author,
                'state': pkg.state,
                'podrocje': podrocje
                }
        pkg_dicts.append(pkg_dict)
    return {'table': pkg_dicts}

unpublished_report_info = {
    'name': 'unpublished',
    'title': u'Zbirke v potrjevanju',
    'description': u'Seznam osnutkov, ki so jih avtorji poslali glavnim urednikom in čakajo na potrditev za objavo.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': unpublished,
    'template': 'report/unpublished.html',
    }


def draft():
    pkgs = model.Session.query(model.Package)\
                .filter(model.Package.state=='draft')\
                .join(model.PackageExtra)\
                .filter_by(key='unpublished')\
                .filter_by(value='false')\
                .all()
    pkg_dicts = []
    for pkg in pkgs:
        org = pkg.get_organization()
        podrocje = pkg.extras.get('podrocje', u'')
        author = dgu_helpers.get_user_name(pkg.creator_user_id)
        pkg_dict = {
                'id': pkg.id,
                'name': pkg.name,
                'title': pkg.title,
                'organization title': org.title if org else u"",
                'organization name': org.name if org else u"",
                'notes': pkg.notes,
                'author': author,
                'state': pkg.state,
                'podrocje': podrocje
                }
        pkg_dicts.append(pkg_dict)
    return {'table': pkg_dicts}

draft_report_info = {
    'name': 'draft',
    'title': u'Zbirke v osnutku',
    'description': u'Seznam osnutkov, ki so jih avtorji poslali glavnim urednikom in čakajo na potrditev za objavo.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': draft,
    'template': 'report/draft.html',
    }

def last_resource_deleted(pkg):

    resource_revisions = model.Session.query(model.ResourceRevision) \
                              .join(model.ResourceGroup) \
                              .join(model.Package) \
                              .filter_by(id=pkg.id) \
                              .order_by(model.ResourceRevision.revision_timestamp) \
                              .all()
    previous_rr = None
    # go through the RRs in reverse chronological order and when an active
    # revision is found, return the rr in the previous loop.
    for rr in resource_revisions[::-1]:
        if rr.state == 'active':
            return previous_rr.revision_timestamp, previous_rr.url
        previous_rr = rr
    return None, ''

def datasets_without_resources():
    pkg_dicts = []
    pkgs = model.Session.query(model.Package)\
                .filter_by(type='dataset') \
                .filter_by(state='active')\
                .order_by(model.Package.title)\
                .all()
    for pkg in pkgs:
        if len(pkg.resources) != 0 or \
          pkg.extras.get('unpublished', '').lower() == 'true':
            continue
        org = pkg.get_organization()
        deleted, url = last_resource_deleted(pkg)
        pkg_dict = OrderedDict((
                ('name', pkg.name),
                ('title', pkg.title),
                ('organization title', org.title),
                ('organization name', org.name),
                ('metadata created', pkg.metadata_created.isoformat()),
                ('metadata modified', pkg.metadata_modified.isoformat()),
                ('last resource deleted', deleted.isoformat() if deleted else None),
                ('last resource url', url),
                ('dataset_notes', lib.dataset_notes(pkg)),
                ))
        pkg_dicts.append(pkg_dict)
    return {'table': pkg_dicts}


datasets_without_resources_info = {
    'name': 'datasets-without-resources',
    'title': u'Zbirke brez datotek',
    'description': u'Zbirke, ki nimajo objavljenih nobenih datotek ali povezav. To ne vključuje osnutkov.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': datasets_without_resources,
    'template': 'report/datasets_without_resources.html',
    }

# app-dataset

def app_dataset_report():
    app_dataset_dicts = []
    for related in model.Session.query(model.RelatedDataset) \
                        .filter(model.Related.type=='App') \
                        .all():
        dataset = related.dataset
        org = dataset.get_organization()
        top_org = list(go_up_tree(org))[-1]

        app_dataset_dict = OrderedDict((
            ('app title', related.related.title),
            ('app url', related.related.url),
            ('dataset name', dataset.name),
            ('dataset title', dataset.title),
            ('organization title', org.title),
            ('organization name', org.name),
            ('top-level organization title', top_org.title),
            ('top-level organization name', top_org.name),
            ('dataset theme', related.dataset.extras.get('theme-primary', '')),
            ('dataset notes', lib.dataset_notes(dataset)),
            ))
        app_dataset_dicts.append(app_dataset_dict)

    app_dataset_dicts.sort(key=lambda row: row['top-level organization title']
                           + row['organization title'])

    return {'table': app_dataset_dicts}

app_dataset_report_info = {
    'name': 'app-dataset-report',
    'title': 'Apps with datasets',
    'description': 'Datasets that have been used by apps.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': app_dataset_report,
    'template': 'report/app_dataset.html',
    }

# app-dataset by theme

def app_dataset_theme_report():
    table = []

    datasets = collections.defaultdict(lambda: {'apps': []})
    for related in model.Session.query(model.RelatedDataset).filter(model.Related.type=='App').all():
        dataset_name = related.dataset.name

        app = {
          'title': related.related.title,
          'url': related.related.url
        }

        datasets[dataset_name]['title'] = related.dataset.title
        datasets[dataset_name]['theme'] = related.dataset.extras.get('theme-primary', '')
        datasets[dataset_name]['apps'].append(app)

    for dataset_name, dataset in datasets.items():
        sorted_apps = sorted(dataset['apps'], key=lambda x: x['title'])
        table.append({'dataset_title': dataset['title'],
                      'dataset_name': dataset_name,
                      'theme': dataset['theme'],
                      'app_titles': "\n".join(a['title'] for a in sorted_apps),
                      'app_urls': "\n".join(a['url'] for a in sorted_apps)})

    return {'table': table}

app_dataset_theme_report_info = {
    'name': 'app-dataset-theme-report',
    'title': 'Apps with datasets by theme',
    'description': 'Datasets that have been used by apps, grouped by theme.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': app_dataset_theme_report,
    'template': 'report/app_dataset_theme_report.html',
    }

# admin-editor report

def get_user_realname(user):
    from ckanext.dgu.drupalclient import DrupalClient
    from HTMLParser import HTMLParser

    if len(user.id) < 10:
        drupal_id = user.id

        html_parser = HTMLParser()

        try:
            dc = DrupalClient()
            properties = dc.get_user_properties(drupal_id)
        except Exception, ex:
            return user.fullname

        try:
            first_name = properties['field_first_name']['und'][0]['safe_value']
            first_name = html_parser.unescape(first_name)
        except:
            first_name = ''

        try:
            surname = properties['field_surname']['und'][0]['safe_value']
            surname = html_parser.unescape(surname)
        except:
            surname = ''
    else:
        first_name = ''
        surname = ''

    name = '%s %s' % (first_name, surname)
    if name.strip() == '':
        name = user.fullname

    return name

def admin_editor(org=None, include_sub_organizations=False):
    from ckanext.dgu.lib.helpers import group_get_users

    table = []

    if org:
        q = model.Group.all('organization')
        parent = model.Group.by_name(org)
        if not parent:
            raise p.toolkit.ObjectNotFound('Publisher not found')

        if include_sub_organizations:
            child_ids = [ch[0] for ch in parent.get_children_group_hierarchy(type='organization')]
        else:
            child_ids = []

        q = q.filter(model.Group.id.in_([parent.id] + child_ids))

        for g in q.all():
            record = {}
            record['publisher_name'] = g.name
            record['publisher_title'] = g.title

            admin_users = group_get_users(g, capacity='admin')
            admins = []
            for u in admin_users:
                name = get_user_realname(u)
                admins.append('%s <%s>' % (name, u.email))

            record['admins'] = "\n".join(admins)

            editor_users = group_get_users(g, capacity='editor')
            editors = []
            for u in editor_users:
                name = get_user_realname(u)
                editors.append('%s <%s>' % (name, u.email))

            record['editors'] = "\n".join(editors)
            table.append(record)
    else:
        table.append({})

    return {'table': table}

def admin_editor_combinations():
    from ckanext.dgu.lib.helpers import organization_list

    for org, _ in organization_list(top=False):
        yield {'org': org}

def user_is_admin(user, org=None):
    import ckan.lib.helpers as helpers
    if org:
        return helpers.check_access('organization_update', {'id': org.id})
    else:
        # Are they admin of any org?
        return len(user.get_groups('organization', capacity='admin')) > 0

def user_is_rm(user, org=None):
    from pylons import config
    from ast import literal_eval
    from ckanext.dgu.lib.publisher import go_up_tree

    relationship_managers = literal_eval(config.get('dgu.relationship_managers', '{}'))

    allowed_orgs = relationship_managers.get(user.name, [])

    if org:
        for o in go_up_tree(org):
            if o.name in allowed_orgs:
                return True

        return False
    else:
        # Are they RM of any org?
        return len(allowed_orgs) > 0

def admin_editor_authorize(user, options):
    if not user:
        return False

    if user.sysadmin:
        return True

    if options.get('org', False):
        org_name = options["org"]
        org = model.Session.query(model.Group) \
                   .filter_by(name=org_name) \
                   .first()
        if not org:
            return False

        if user_is_admin(user, org) or user_is_rm(user, org):
            return True
        else:
            return False
    else:
        # Allow them to see front page / see report on report index
        if user_is_admin(user) or user_is_rm(user):
            return True

    return False

admin_editor_info = {
    'name': 'admin_editor',
    'title': u'Skrbniki in uredniki organizacij',
    'description': u'Seznam organizacij, ki prikazuje njihove skrbnike in urednike.',
    'option_defaults': OrderedDict((('org', ''), )),
    'option_combinations': admin_editor_combinations,
    'generate': admin_editor,
    'template': 'report/admin_editor.html',
    'authorize': admin_editor_authorize
    }


# LA Schemas


def la_schemas(local_authority=None, schema=None, incentive_only=False):
    from ckanext.dgu.bin.schema_apply_lga import LaSchemas
    Options = collections.namedtuple('Options', ('organization', 'incentive_only', 'schema', 'write', 'dataset', 'print_'))
    options = Options(organization=None, incentive_only=incentive_only,
                      schema=schema, write=False, dataset=None, print_=False)
    csv_filepath = os.path.abspath(os.path.join(__file__, '../../incentive.csv'))
    return LaSchemas.command(config_ini=None, options=options,
                             submissions_csv_filepath=csv_filepath)


def la_schemas_combinations():
    for schema in [''] + dgu_helpers.get_la_schema_options():
        for incentive_only in (False, True):
            yield {'schema': schema['title'] if schema else '',
                   'incentive_only': incentive_only}

la_schemas_info = {
    'name': 'la-schemas',
    'title': 'Schemas for local authorities',
    'description': 'Schemas matched to local authority datasets.',
    'option_defaults': OrderedDict((('schema', ''),
                                    ('incentive_only', False))),
    'option_combinations': la_schemas_combinations,
    'generate': la_schemas,
    'template': 'report/la_schemas.html',
    }


# Licence report

def licence_report(organization=None, include_sub_organizations=False):
    '''
    Returns a dictionary detailing licences for datasets in the
    organisation specified, and optionally sub organizations.
    '''
    # Get packages
    if organization:
        top_org = model.Group.by_name(organization)
        if not top_org:
            raise p.toolkit.ObjectNotFound('Publisher not found')

        if include_sub_organizations:
            orgs = lib.go_down_tree(top_org)
        else:
            orgs = [top_org]
        pkgs = set()
        for org in orgs:
            org_pkgs = model.Session.query(model.Package)\
                            .filter_by(state='active')
            org_pkgs = lib.filter_by_organizations(
                org_pkgs, organization,
                include_sub_organizations=False)\
                .all()
            pkgs |= set(org_pkgs)
    else:
        pkgs = model.Session.query(model.Package)\
                    .filter_by(state='active')\
                    .all()

    # Get their licences
    packages_by_licence = collections.defaultdict(list)
    rows = []
    num_pkgs = 0
    for pkg in pkgs:
        if pkg.state == 'draft':
            # Ignore draft datasets
            continue
        licence_tuple = (pkg.license_id or '',
                         pkg.license.title if pkg.license else '',
                         pkg.extras.get('licence', ''))
        packages_by_licence[licence_tuple].append((pkg.name, pkg.title))
        num_pkgs += 1

    for licence_tuple, dataset_tuples in sorted(packages_by_licence.items(),
                                                key=lambda x: -len(x[1])):
        license_id, license_title, licence = licence_tuple
        dataset_tuples.sort(key=lambda x: x[0])
        dataset_names, dataset_titles = zip(*dataset_tuples)
        licence_dict = OrderedDict((
            ('license_id', license_id),
            ('license_title', license_title),
            ('licence', licence),
            ('dataset_titles', '|'.join(t for t in dataset_titles)),
            ('dataset_names', ' '.join(dataset_names)),
            ))
        rows.append(licence_dict)

    return {
        'num_datasets': num_pkgs,
        'num_licences': len(rows),
        'table': rows,
        }


def licence_combinations():
    for organization in lib.all_organizations(include_none=True):
        for include_sub_organizations in (False, True):
            yield {'organization': organization}


licence_report_info = {
    'name': 'licence',
    'title': u'Licence',
    'description': u'Licence po zbirkah podatkov.',
    'option_defaults': OrderedDict((('organization', None),)),
    'option_combinations': licence_combinations,
    'generate': licence_report,
    'template': 'report/licence_report.html',
    }


# Datasets only in PDF

def pdf_datasets_report():
    '''
    Returns datasets that have data in PDF format, by organization.
    '''
    # Get packages
    pkgs = model.Session.query(model.Package)\
                .filter_by(state='active')

    # See if PDF
    num_datasets_published = 0
    num_datasets_only_pdf = 0
    datasets_by_publisher_only_pdf = collections.defaultdict(list)
    # use yield_per, otherwise memory use just goes up til the script is killed
    # by the os.
    for pkg in pkgs.yield_per(100):
        if pkg.state == 'draft':
            continue
        num_datasets_published += 1

        formats = set([res.format.lower() for res in pkg.resources
                       if res.resource_type != 'documentation'])
        if 'pdf' not in formats:
            continue
        org = pkg.get_organization().name

        data_formats = formats - set(('html', '', None))
        if data_formats == set(('pdf',)):
            num_datasets_only_pdf += 1
            datasets_by_publisher_only_pdf[org].append((pkg.name, pkg.title))

    rows = []
    for org_name, datasets_only_pdf in sorted(
            datasets_by_publisher_only_pdf.iteritems(),
            key=lambda x: -len(x[1])):
        org = model.Session.query(model.Group) \
                   .filter_by(name=org_name) \
                   .first()
        top_org = list(go_up_tree(org))[-1]

        row = OrderedDict((
            ('organization title', org.title),
            ('organization name', org.name),
            ('top-level organization title', top_org.title),
            ('top-level organization name', top_org.name),
            ('num datasets only pdf', len(datasets_only_pdf)),
            ('name datasets only pdf',
             ' '.join(d[0] for d in datasets_only_pdf)),
            ('title datasets only pdf',
             '|'.join(d[1] for d in datasets_only_pdf)),
            ))
        rows.append(row)

    return {'table': rows,
            'num_datasets_published': num_datasets_published,
            'num_datasets_only_pdf': num_datasets_only_pdf,
            }


pdf_datasets_report_info = {
    'name': 'pdf_datasets',
    'title': 'PDF Datasets',
    'description': 'Datasets with data only in PDF format.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': pdf_datasets_report,
    'template': 'report/pdf_datasets_report.html',
    }


# Datasets with HTML link

def html_datasets_report():
    '''
    Returns datasets that only have an HTML link, by organization.
    '''
    # Get packages
    pkgs = model.Session.query(model.Package)\
                .filter_by(state='active')

    # See if HTML
    num_datasets_published = 0
    num_datasets_only_html = 0
    datasets_by_publisher_only_html = collections.defaultdict(list)
    # use yield_per, otherwise memory use just goes up til the script is killed
    # by the os.
    for pkg in pkgs.yield_per(100):
        if pkg.state == 'draft':
            continue
        num_datasets_published += 1

        formats = set([res.format.lower() for res in pkg.resources
                       if res.resource_type != 'documentation'])
        if 'html' not in formats:
            continue
        org = pkg.get_organization().name

        data_formats = formats - set(('asp', '', None))
        if data_formats == set(('html',)):
            num_datasets_only_html += 1
            datasets_by_publisher_only_html[org].append((pkg.name, pkg.title))

    rows = []
    for org_name, datasets_only_html in sorted(
            datasets_by_publisher_only_html.iteritems(),
            key=lambda x: -len(x[1])):
        org = model.Session.query(model.Group) \
                   .filter_by(name=org_name) \
                   .first()
        top_org = list(go_up_tree(org))[-1]

        row = OrderedDict((
            ('organization title', org.title),
            ('organization name', org.name),
            ('top-level organization title', top_org.title),
            ('top-level organization name', top_org.name),
            ('num datasets only html', len(datasets_only_html)),
            ('name datasets only html',
             ' '.join(d[0] for d in datasets_only_html)),
            ('title datasets only html',
             '|'.join(d[1] for d in datasets_only_html)),
            ))
        rows.append(row)

    return {'table': rows,
            'num_datasets_published': num_datasets_published,
            'num_datasets_only_html': num_datasets_only_html,
            }


html_datasets_report_info = {
    'name': 'html_datasets',
    'title': 'HTML Datasets',
    'description': 'Datasets with data only a link to an HTML page.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': html_datasets_report,
    'template': 'report/html_datasets_report.html',
    }


completeness_main_fields = ['title', 'notes', 'tags', 'owner_org', 'metadata_created', 'metadata_modified', 'version']
completeness_extra_fields = ['podrocje', 'jezik_zbirke', 'contact-name', 'contact-email', 'contact-phone',
                'foi-name', 'foi-email', 'foi-phone', 
                'licence', 'pogostost_osvezevanja', 'open', 'pogoji_licence', 'pogoji_licence_cena', 'dostopnost_zbirke',
                'sop_stevilka', 'vrsta_pravne_podlage', 'povezanost_desc',
                'temporal_coverage-from', 'temporal_coverage-to', 'geographic_coverage']
completeness_resource_fields = ['resource_access_url', 'resource_download_url', 'resource_format', 'resource_description', 'resource_licence']
completeness_special_fields = ['contains_resources', 'contains_file_resources', 'contains_additional_resources', 'pravne_podlage']

def retrieve_all_packages():
    return model.Session.query(model.Package)\
                .filter_by(state='active')\
                .order_by(model.Package.title)\
                .all()

# list packages that have a specific field missing or null
def list_incomplete_packages(missing_field, as_urls=False, pkgs=None):
    if not pkgs:
        pkgs=retrieve_all_packages()

    result_list = []
    for package in pkgs:
        if missing_field in completeness_main_fields:
            if missing_field == 'tags':
                if not package.get_tags():
                    result_list.append(package.name)
            elif not getattr(package, missing_field, None):
                result_list.append(package.name)
        elif missing_field in completeness_extra_fields:    
            if not package.extras.get(missing_field, None):
                result_list.append(package.name)
        elif missing_field == 'pravne_podlage':
            if not package.extras.get('sop_stevilka', None) and not package.extras.get('vrsta_pravne_podlage', None):
                result_list.append(package.name)
        elif missing_field in completeness_special_fields:
            # we need to process resources
            if package.resources and len(package.resources) > 0:
                file_found = False
                docs_found = False
                non_docs_found = False
                for res in package.resources:
                    # check if documentation
                    if res.resource_type == 'documentation':
                        docs_found = True
                        continue
                    non_docs_found = True
                            
                    # check if it is not a webpage - if not, we presume it is a file
                    format_webpage = False
                    if res.format.lower() == 'html' or res.format.lower() == 'aspx':
                        format_webpage = True 
                    url_webpage = False
                    if res.url.lower().endswith('html') or res.url.lower().endswith('aspx'):
                        url_webpage = True
                    if not format_webpage and not url_webpage:
                        file_found = True
                if missing_field == 'contains_file_resources':
                    if not file_found:
                        result_list.append(package.name)
                if missing_field == 'contains_additional_resources':
                    if not docs_found:
                        result_list.append(package.name)
                if missing_field == 'contains_resources':
                    if not non_docs_found:
                        result_list.append(package.name)
            else:
                result_list.append(package.name)
        else: #unknown field name
            return []
    if as_urls:
        result_list = [url_for(controller='ckanext.dgu.controllers.package:PackageController', action='read', qualified=True, id=result) for result in result_list]
    return result_list
        
def list_incomplete_resources(missing_field, as_urls=False, pkgs=None):
    if not pkgs:
        pkgs=retrieve_all_packages()

    result_list = []
    if missing_field not in completeness_resource_fields:
        return [] #not resource metric
    for package in pkgs:
        if package.resources and len(package.resources) > 0:
            for res in package.resources:
                # check if documentation
                if res.resource_type == 'documentation':
                    continue

                if missing_field == 'resource_access_url':
                    if not getattr(res, 'url', None):
                        result_list.append((package.id, res.id)) 
                elif missing_field == 'resource_format':
                    if not getattr(res, 'format', None):
                        result_list.append((package.id, res.id))
                elif missing_field == 'resource_description':
                    if not getattr(res, 'description', None):
                        result_list.append((package.id, res.id))
                elif missing_field == 'resource_licence':
                    if not package.extras.get('licence', None):
                        result_list.append((package.id, res.id))
                elif missing_field == 'resource_download_url':
                    # check if it is not a webpage - if not, we presume it is a file
                    format_webpage = False
                    if res.format.lower() == 'html' or res.format.lower() == 'aspx':
                        format_webpage = True 
                    url_webpage = False
                    if res.url.lower().endswith('html') or res.url.lower().endswith('aspx'):
                        url_webpage = True
                    if format_webpage or url_webpage:
                        result_list.append((package.id, res.id))
    if as_urls:
        result_list = [url_for(controller='ckanext.dgu.controllers.package:PackageController', action='resource_read', qualified=True, id=result[0], resource_id=result[1]) for result in result_list]
    else:
        result_list = [result[1] for result in result_list]
    return result_list
                
def get_data_resource_num(pkgs):
    data_resource_num = 0
    for package in pkgs:            
        if package.resources and len(package.resources) > 0:
            for res in package.resources:
                # do not count documentation
                if res.resource_type != 'documentation':
                    data_resource_num = data_resource_num + 1
    return data_resource_num

def completeness_report():
    package_table = []
    resource_table = []

    pkgs = retrieve_all_packages()

    main_fields = ['title', 'notes', 'tags', 'owner_org', 'metadata_created', 'metadata_modified', 'version']
    extra_fields = ['podrocje', 'jezik_zbirke', 'contact-name', 'contact-email', 'contact-phone',
                    'foi-name', 'foi-email', 'foi-phone', 
                    'licence', 'pogostost_osvezevanja', 'open', 'pogoji_licence', 'pogoji_licence_cena', 'dostopnost_zbirke',
                    'sop_stevilka', 'vrsta_pravne_podlage', 'povezanost_desc',
                    'temporal_coverage-from', 'temporal_coverage-to', 'geographic_coverage']
    resource_fields = ['resource_access_url', 'resource_download_url', 'resource_format', 'resource_description', 'resource_licence']
    special_fields = ['contains_resources', 'contains_file_resources', 'contains_additional_resources', 'pravne_podlage']
    
    # mappings from variable to user-friendly name, DCAT mapping and its necessity
    mappings = {'title':                         (u'Naslov zbirke', 'dct:title', 'Mandatory'), 
                'notes':                         (u'Opis zbirke', 'dct:description', 'Mandatory'), 
                'tags':                          (u'Ključne besede', 'dcat:keyword', 'Recommended'), 
                'owner_org':                     (u'Organizacija', 'dct:publisher', 'Recommended'), 
                'metadata_created':              (u'Datum objave', 'dct:issued', 'Optional'), 
                'metadata_modified':             (u'Datum posodobitve', 'dct:modified', 'Optional'), 
                'version':                       (u'Verzija', 'owl:versionInfo', 'Optional'),
                'podrocje':                      (u'Področje', 'dcat:theme', 'Recommended'), 
                'jezik_zbirke':                  (u'Jezik zbirke', 'dct:language', 'Optional'), 
                'contact-name':                  (u'Organizacija - ime', 'dct:publisher - foaf:name', 'Recommended'), 
                'contact-email':                 (u'Organizacija - e-pošta', '/', '/'), 
                'contact-phone':                 (u'Organizacija - telefon', '/', '/'),
                'foi-name':                      (u'Skrbnik - ime', 'dcat:contactPoint - vcard:fn', 'Recommended'), 
                'foi-email':                     (u'Skrbnik - e-pošta', 'dcat:contactPoint - vcard:hasEmail', 'Optional'), 
                'foi-phone':                     (u'Skrbnik - telefon', 'dcat:contactPoint - vcard:hasTelephone', 'Optional'), 
                'licence':                       (u'Licenca', 'dcat:distribution - dct:license', 'Recommended'), 
                'pogostost_osvezevanja':         (u'Pogostost osveževanja', 'dct:accrualPeriodicity', 'Optional'), 
                'open':                          (u'Odprta licenca', '/', '/'), 
                'pogoji_licence':                (u'Pogoji licence', '/', '/'), 
                'pogoji_licence_cena':           (u'Pogoji licence - cena', '/', '/'), 
                'dostopnost_zbirke':             (u'Dostopnost zbirke', 'dct:accessRights', 'Optional'),
                'sop_stevilka':                  (u'Pravne podlage - SOP', '/', '/'), 
                'vrsta_pravne_podlage':          (u'Pravne podlage - ne SOP', '/', '/'),
                'pravne_podlage':                (u'Pravne podlage - skupno', 'dct:conformsTo', 'Optional'), 
                'povezanost_desc':               (u'Povezanost z drugimi viri', 'dct:relation', 'Optional'),
                'temporal_coverage-from':        (u'Časovni obseg - od', 'dct:temporal', 'Optional'), 
                'temporal_coverage-to':          (u'Časovni obseg - do', 'dct:temporal', 'Optional'), 
                'geographic_coverage':           (u'Geografsko področje', 'dct:spatial', 'Optional'),
                'contains_resources':            (u'Vsebuje povezave', 'dcat:distribution - dcat:accessURL', 'Recommended'), 
                'contains_file_resources':       (u'Vsebuje podatkovne povezave', 'dcat:distribution - dcat:downloadURL', 'Recommended'), 
                'contains_additional_resources': (u'Vsebuje dodatne povezave', 'foaf:page', 'Optional'),
                'resource_access_url':           (u'Povezava na spletno stran', 'dcat:distribution - dcat:accessURL', 'Mandatory'),
                'resource_download_url':         (u'Povezava do datoteke', 'dcat:distribution - dcat:downloadURL', 'Optional'),
                'resource_format':               (u'Povezava ima format', 'dcat:distribution - dct:format', 'Recommended'),
                'resource_description':          (u'Povezava ima opis', 'dcat:distribution - dct:description', 'Recommended'),
                'resource_licence':              (u'Povezava ima licenco (zbirke)', 'dcat:distribution - dct:license', 'Recommended')}
                
    counts = collections.OrderedDict()
    
    for field in completeness_main_fields + completeness_extra_fields + completeness_special_fields:
        counts[field] = len(pkgs) - len(list_incomplete_packages(field, False, pkgs))
        
    resource_counts = {}
    data_resource_num = get_data_resource_num(pkgs)
    for field in resource_fields:
        resource_counts[field] = data_resource_num - len(list_incomplete_resources(field, False, pkgs))

    table_by_necessity = {'Mandatory': [], 'Recommended': [], 'Optional': [], '/': []}
    for field_name, package_count in counts.items():
        link_to_api = ""
        if package_count < len(pkgs):
            link_to_api = url_for(controller='ckanext.dgu.controllers.api:DguApiController', action='list_incomplete_packages', qualified=True, missing_field=field_name, as_urls=True)
        table_by_necessity[mappings[field_name][2]].append({'field': mappings[field_name][0],
                      'dcat-mapping': mappings[field_name][1],
                      'dcat-necessity': mappings[field_name][2],
                      'number': package_count,
                      'percentage': round(package_count*10000.0/len(pkgs))/100.0,
                      'missing-list': link_to_api})
    table_by_necessity['Mandatory'] = sorted(table_by_necessity['Mandatory'], key=lambda k: k['number'], reverse=True) 
    table_by_necessity['Recommended'] = sorted(table_by_necessity['Recommended'], key=lambda k: k['number'], reverse=True) 
    table_by_necessity['Optional'] = sorted(table_by_necessity['Optional'], key=lambda k: k['number'], reverse=True)
    table_by_necessity['/'] = sorted(table_by_necessity['/'], key=lambda k: k['number'], reverse=True)
    package_table = table_by_necessity['Mandatory'] + table_by_necessity['Recommended'] + table_by_necessity['Optional'] + table_by_necessity['/']
    
    table_by_necessity = {'Mandatory': [], 'Recommended': [], 'Optional': [], '/': []}    
    for field_name, resource_count in resource_counts.items():
        link_to_api = ""
        if resource_count < data_resource_num:
            link_to_api = url_for(controller='ckanext.dgu.controllers.api:DguApiController', action='list_incomplete_resources', qualified=True, missing_field=field_name, as_urls=True)
        table_by_necessity[mappings[field_name][2]].append({'field': mappings[field_name][0],
                      'dcat-mapping': mappings[field_name][1],
                      'dcat-necessity': mappings[field_name][2],
                      'number': resource_count,
                      'percentage': round(resource_count*10000.0/data_resource_num)/100.0,
                      'missing-list': link_to_api})
    table_by_necessity['Mandatory'] = sorted(table_by_necessity['Mandatory'], key=lambda k: k['number'], reverse=True) 
    table_by_necessity['Recommended'] = sorted(table_by_necessity['Recommended'], key=lambda k: k['number'], reverse=True) 
    table_by_necessity['Optional'] = sorted(table_by_necessity['Optional'], key=lambda k: k['number'], reverse=True)
    table_by_necessity['/'] = sorted(table_by_necessity['/'], key=lambda k: k['number'], reverse=True)
    resource_table = table_by_necessity['Mandatory'] + table_by_necessity['Recommended'] + table_by_necessity['Optional'] + table_by_necessity['/']

    return {'table': package_table + resource_table, 'package_table': package_table, 'package_num': len(pkgs), 'data_resource_table': resource_table, 'data_resource_num': data_resource_num}

completeness_report_info = {
    'name': 'completeness',
    'title': u'Izpolnjena polja v zbirkah',
    'description': u'Deleži, koliko zbirk ima izpolnjena določena metapodatkovna polja.',
    'option_defaults': None,
    'option_combinations': None,
    'generate': completeness_report,
    'template': 'report/completeness_report.html',
    }


