import logging

import ckan.logic as logic
from ckan import model
from ckan.common import _

from ckanext.dgu.model.publisher_request import PublisherRequest

log = logging.getLogger(__name__)


get_action = logic.get_action

# a couple of users have different mails on multiple accounts
# provide proper exceptions for these by specifying mappings
# from their email hashes to the ID of matching duplicate user
merge_direct = {'640e8c7cd5503bfea12f971b16f8a32d': '192',
                '35722069136f534c9e6210d3982bcf9a': '121',
                '0381caca8377f2de514bf5e6b056b900': '184',
                'b2a66a3026b605ed8169d2200338b713': '265'
    }

def merge_user (user, original_name):
    try:
        log.debug('Merging user ...')
        if not isinstance(user, model.User):
            log.error('Could not merge user ' + str(user) + ', not a user object!')
            return
        if user.email_hash in merge_direct.keys():
            move_user(user.id, merge_direct.get(user.email_hash()))
        else:
            users = model.Session.query(model.User).filter_by(email=user.email)
            for u in users:
                log.debug('Checking user ' + u.id)
                if u.id != user.id:
                    move_user(u.id, user.id)
        change_username(user, original_name)
    except Exception as e:
        log.error('Error merging user' + str(user), e)
        
def change_username (user, original_name):    
    # update publisher requests of user
    # no revisioning is used for publisher_request
    model.Session.query(PublisherRequest)\
                 .filter(PublisherRequest.user_name == original_name)\
                 .update({"user_name": user.name})
                 
    # update revisions of user
    model.Session.query(model.Revision)\
                 .filter(model.Revision.author == original_name)\
                 .update({"author": user.name})

def move_user (old_id, new_id):
    ''' Migrate ownership of data from a specific user to a new user.
    Basically, this consists of changing references to the 'id' in table 'user'. 
    
    The following tables contain most important data that need to be migrated:
    package (creator_user_id)
    harvest_source (user_id)
    publisher_request(user_name)
    
    The following references just need to be recreated for the new user:
    member & member_revision (table_name==user && table_id)
    user_object_role (user_id)
    
    The following are of lesser importance and will be kept as is:
    *user_following_dataset
    *user_following_group
    *activity (user_id)
    *dashboard (user_id)
    *issue (user_id)
    *issue_comment (user_id)
    *issue_comment_report (user_id)
    *rating (user_id)
    
    The following are deprecated or historical and will be ignored
    *authorization_group_user (user_id)
    package_revision (creator_user_id)'''

    log.debug('Moving user ' + old_id + ' to ' + new_id)
    
    # get new user 
    new_user = model.User.get(new_id)
    if not isinstance(new_user, model.User):
        log.error('Could not migrate data from ' + str(old_id) + ' to ' + str(new_id) + ', new user not found!')
        return
    
    # get old user
    old_user = model.User.get(old_id)
    if not isinstance(old_user, model.User):
        log.error('Could not migrate data from ' + str(old_id) + ' to ' + str(new_id) + ', old user not found!')
        return
    
    # get packages of old user
    packages = model.Session.query(model.Package)\
            .filter(model.Package.creator_user_id==old_id)
    
    # update all found packages with new creator id
    for pkg in packages:
        pkg.creator_user_id=new_id

        rev = model.repo.new_revision() 
        rev.author = u'user-migration'
        rev.message = u'Merged users after change of drupal usernames'

        model.Session.commit()
        log.debug("Moved package " + str(pkg.name) + " to user " + str(new_user.name))
        
    # get harvest sources of old user
    try:
        from ckanext.harvest.model import HarvestSource
        
        # update all found harvest sources with new creator id
        # no revisioning is used for harvest sources
        harvest_sources = model.Session.query(HarvestSource)\
            .filter(HarvestSource.user_id == old_id).update({"user_id": new_id})
        
    except ImportError:
        log.error('Could not migrate user for harvest sources as harvest plugin could not be imported')
        #do nothing
    
    # update publisher requests of old user
    # no revisioning is used for publisher_request
    model.Session.query(PublisherRequest)\
                 .filter(PublisherRequest.user_name == old_user.name)\
                 .update({"user_name": new_user.name})
                 
    # update revisions of old user
    model.Session.query(model.Revision)\
                 .filter(model.Revision.author == old_user.name)\
                 .update({"author": new_user.name})             
    
    # get old user memberships
    memberships = model.Session.query(model.Member)\
                        .filter(model.Member.state == 'active')\
                        .filter(model.Member.table_name == 'user')\
                        .filter(model.Member.table_id == old_id)
    existing_groups = new_user.get_groups()
   
    # copy old user memberships to new user
    for m in memberships:
        already_member = False
        for g in existing_groups:
            if g.id == m.group_id:
                log.debug("Merged user already part of group " + str(m.group_id))
                already_member = True
                break
        if not already_member:
            log.debug("Adding membership " + str(m) + " to user " + str(new_id))
            member_obj = model.Member(table_id=new_id,
                                      table_name='user',
                                      capacity=m.capacity,
                                      group_id=m.group_id,
                                      state='active')
            model.Session.add(member_obj)
    
    # get old user group roles
    roles = model.Session.query(model.GroupRole)\
                         .filter(model.GroupRole.user == old_user)
    existing_roles = model.Session.query(model.GroupRole)\
                                  .filter(model.GroupRole.user == new_user)
  
    # copy old user group roles to new user
    for role in roles:
        already_member = False
        for r in existing_roles:
            if r.role == role.role and r.group == role.group:
                log.debug("Merged user already has role " + str(r.role) + " for group " + str(r.group))
                already_member = True
                break
        if not already_member:
            log.debug("Adding group role " + str(role) + " to user " + str(new_id))
            role_obj = model.GroupRole(user=new_user,
                                       role=role.role,
                                       group=role.group)
            model.Session.add(role_obj)
    
    # get old user package roles
    roles = model.Session.query(model.PackageRole)\
                         .filter(model.PackageRole.user == old_user)
    existing_roles = model.Session.query(model.PackageRole)\
                                  .filter(model.PackageRole.user == new_user)
    # copy old user package roles to new user
    for role in roles:
        already_member = False
        for r in existing_roles:
            if r.role == role.role and r.package == role.package:
                log.debug("Merged user already has role " + str(r.role) + " for package " + str(r.package))
                already_member = True
                break
        if not already_member:
            log.debug("Adding package role " + str(role) + " to user " + str(new_id))
            role_obj = model.PackageRole(user=new_user,
                                         role=role.role,
                                         package=role.package)
            model.Session.add(role_obj)
    
    log.debug('Merged user ' + old_id + ' to ' + new_id)


def merge_users_on_delete(user_id, new_user_id, context):
    result = {}
    new_user = model.User.get(new_user_id)

    if not isinstance(new_user, model.User):
        log.error('Could not migrate data from ' + str(user_id) + ' to ' + str(new_user_id) + ', new user not found!')
        result['error'] = '%s (%s)' % (_('User not found'), str(new_user_id))
        return result

    packages = model.Session.query(model.Package) \
        .filter(model.Package.creator_user_id == user_id)

    for pkg in packages:
        role_show_dict = {'domain_object': pkg.id, 'roles': ['admin']}
        existing_roles = logic.get_action('roles_show')(context, role_show_dict)
        role_update_dict_list = [{'user': new_user.id, 'roles': ['admin']}]
        for role in existing_roles.get('roles', []):
            if role.get('role', '') == 'admin' and role.get('user_id', '') != new_user:
                role_update_dict_list.append({'user': role.get('user_id'), 'roles': []})
        updated_roles = logic.get_action('user_role_bulk_update')(context, {'domain_object': pkg.id, 'user_roles': role_update_dict_list})
    return result
