# -*- coding: utf-8 -*-

import ckan.lib.uploader as uploader
import ckan.lib.navl.dictization_functions
import ckan.logic as logic
import logging

from ckanext.dgu.lib.reports import refresh_opsi_reports_async
from ckanext.dgu.forms.dataset_form import DatasetForm, convert_to_extras
from ckanext.dgu.plugins_toolkit import _
log = logging.getLogger(__name__)
_validate = ckan.lib.navl.dictization_functions.validate
_get_action = logic.get_action
_check_access = logic.check_access
NotFound = logic.NotFound
ValidationError = logic.ValidationError
_get_or_bust = logic.get_or_bust

def opsi_action_resource_update(context, data_dict):
    '''Update a resource.

    To update a resource you must be authorized to update the dataset that the
    resource belongs to.

    For further parameters see ``resource_create()``.

    :param id: the id of the resource to update
    :type id: string

    :returns: the updated resource
    :rtype: string

    '''
    model = context['model']
    user = context['user']
    id = _get_or_bust(data_dict, "id")

    resource = model.Resource.get(id)
    context["resource"] = resource

    if not resource:
        raise NotFound(_('Resource was not found.'))

    _check_access('resource_update', context, data_dict)
    del context["resource"]

    package_id = resource.resource_group.package.id
    # DGU-only hack - http://redmine.dguteam.org.uk/issues/1281
    # Provide a local context, because package_show will set a schema in it
    # that is only appropriate for package_show.
    # OPSI: removing this hack, because it interferes with resource_update and our schema 
    # (Missing validation errors, because default schema does not contain all mandatory fields
    #package_show_context = {'model': model, 'user': user}
    #pkg_dict = _get_action('package_show')(dict(package_show_context, return_type='dict'),
    #                                       {'id': package_id,
    #                                        'use_default_schema': True})
    pkg_dict = _get_action('package_show')(dict(context, return_type='dict', for_update=True),{'id': package_id})

    for n, p in enumerate(pkg_dict['resources']):
        if p['id'] == id:
            break
    else:
        log.error('Could not find resource %s after all', id)
        raise NotFound(_('Resource was not found.'))

    # Persist the datastore_active extra if already present and not provided
    if ('datastore_active' in resource.extras and
            'datastore_active' not in data_dict):
        data_dict['datastore_active'] = resource.extras['datastore_active']

    #remove date if it is empty (should be saved as individual resource)
    if data_dict.get('date', 'not existing') == '':
        del data_dict['date']

    #if description is missing, but name is specified, copy it
    if not data_dict.get('description') and data_dict.get('name'):
        data_dict['description'] = data_dict.get('name')

    #pogostost_osvezevanja is override of package field and we remove it from resource
    if data_dict.has_key('pogostost_osvezevanja'):
        if data_dict.get('pogostost_osvezevanja') != '':
            pkg_dict['pogostost_osvezevanja'] = data_dict['pogostost_osvezevanja']
        del data_dict['pogostost_osvezevanja']

    # resource has to be updated in all resource lists,
    # otherwise it would be merged back to the list of resources
    # and single resource could not be updated
    if 'clear_upload' not in data_dict: # if we are not preparing resource for deleting
        if pkg_dict.get('individual_resources'):
            for k, p in enumerate(pkg_dict['individual_resources']):
                if p['id'] == id:
                    pkg_dict['individual_resources'][k] = data_dict
                    break
        if pkg_dict.get('timeseries_resources'):
            for l, p in enumerate(pkg_dict['timeseries_resources']):
                if p['id'] == id:
                    pkg_dict['timeseries_resources'][l] = data_dict
                    break
        if pkg_dict.get('additional_resources'):
            for m, p in enumerate(pkg_dict['additional_resources']):
                if p['id'] == id:
                    pkg_dict['additional_resources'][m] = data_dict
                    break

    upload = uploader.ResourceUpload(data_dict)

    pkg_dict['resources'][n] = data_dict

    try:
        context['defer_commit'] = True
        context['use_cache'] = False
        context['ignore_auth'] = True #content editor will otherwise not be able to upload to active datasets
        pkg_dict = _get_action('package_update')(context, pkg_dict)
        context.pop('defer_commit')
    except ValidationError, e:
        # For debugging #1281
        if ('resources' not in e.error_dict and 'individual_resources' not in e.error_dict and
            'timeseries_resources' not in e.error_dict and 'additional_resources' not in e.error_dict):
            raise Exception('resource_update error: %r' %
                            e.error_dict)

        if 'resources' in e.error_dict:
            errors = e.error_dict['resources'][n]
        elif 'individual_resources' in e.error_dict:
            errors = e.error_dict['individual_resources'][k]
        elif 'timeseries_resources' in e.error_dict:
            errors = e.error_dict['timeseries_resources'][l]
        elif 'additional_resources' in e.error_dict:
            errors = e.error_dict['additional_resources'][m]
        raise ValidationError(errors)

    resource_size_limit = context.get('max_resource_size', 16*1024*1024*1024) if context.get('is_local') else uploader.get_max_resource_size()
    upload.upload(id, resource_size_limit)
    model.repo.commit()
    return _get_action('resource_show')(context, {'id': id})



def opsi_action_bulk_publish(context, data_dict):
    _check_access('bulk_publish', context, data_dict)

    packages = data_dict.get('packages', [])
    model = context['model']
    user = context.get('user')

    rev = model.repo.new_revision()
    rev.author = str(user)

    response = {
        'errors': {},
        'success': []
    }

    context['defer_commit'] = True
    context['use_cache'] = False
    context['ignore_auth'] = True

    for package in packages:
        package_error_key = package.get('name', package['id'])
        try:
            pkg_dict = _get_action('package_show')(dict(context, return_type='dict'), {'id': package['id']})
        except NotFound:
            errors = response['errors'].get(package_error_key, [])
            errors.append('Package does not exists.')
            response['errors'][package_error_key] = errors
            continue
        pkg_dict['podrocje'] = package['podrocje']
        pkg_dict['state'] = 'active'
        pkg_dict['unpublished'] = False
        context['defer_commit'] = True
        try:
            pkg_dict = _get_action('package_update')(context, pkg_dict)
            context.pop('defer_commit')
            model.repo.commit()
            response['success'].append(package_error_key)
        except ValidationError, e:
            errors = response['errors'].get(package_error_key, [])
            for key, value in e.error_dict.items():
                msg = '%s: %s' % (key, ", ".join([v.encode('utf8') for v in value]))
                errors.append(msg)
                response['errors'][package_error_key] = errors
                log.error('Validation error while bulk publish package: ' + str(e))
            continue
        except Exception, e:
            errors = response['errors'].get(package_error_key, [])
            errors.append(u'Pri objavljanju zbirke je prišlo do napake.')
            response['errors'][package_error_key] = errors
            log.error('Error while bulk publish package: ' +str(e))
            continue

    if len(response['success']) > 0:
        refresh_opsi_reports_async()

    return response
