﻿from logging import getLogger

from ckan.lib.helpers import flash_notice
import ckan.plugins as p
import ckan.plugins.toolkit as toolkit
from ckanext.dgu.authentication.drupal_auth import DrupalAuthMiddleware
from ckanext.dgu.lib.site_down_middleware import SiteDownMiddleware
from ckanext.dgu.authorize import (
                             opsi_group_update,
                             opsi_organization_update,
                             opsi_package_change_state,
                             opsi_package_create,
                             opsi_resource_delete,
                             opsi_package_show,
                             opsi_resource_create,
                             opsi_resource_update,
                             opsi_package_update,
                             opsi_extra_fields_editable,
                             opsi_dataset_delete, opsi_user_list, opsi_user_show,
                             dgu_organization_delete, dgu_group_change_state,
                             opsi_bulk_publish,
                             opsi_user_delete
                             )
from ckanext.report.interfaces import IReport
from ckan.lib.helpers import url_for
from ckanext.dgu.lib.helpers import dgu_linked_user, is_plugin_enabled
from ckanext.dgu.search_indexing import SearchIndexing
from ckanext.dgu import gemini_postprocess_tasks
from ckan.config.routing import SubMapper
from ckan.exceptions import CkanUrlException

log = getLogger(__name__)


def after(instance, action, **params):
    from pylons import response
    instance._set_cors()
    response.headers['Vary'] = 'Cookie'

def not_found(self, url):
    from ckan.lib.base import abort
    abort(404)

def _guess_package_type(self, expecting_name=False):
    return 'dataset'

def delete_routes_by_path_startswtih(map, path_startswith):
    matches_to_delete = []
    for match in map.matchlist:
        if match.routepath.startswith(path_startswith):
            matches_to_delete.append(match)
    for match in matches_to_delete:
        map.matchlist.remove(match)

def delete_routes_by_name(map, route_names):
    if isinstance(route_names, basestring):
        route_names = [route_names]
    for route_name in route_names:
        del map._routenames[route_name]


class DguReportPlugin(p.SingletonPlugin):
    p.implements(p.IRoutes, inherit=True)

    def after_map(self, map):
        # Put reports at /data/reports instead of /reports
        # Delete routes to /report otherwise url_for links end up pointed there.
        delete_routes_by_path_startswtih(map, '/report')
        delete_routes_by_name(map, ('reports', 'report', 'report-org'))

        # Add new routes to /data/reports
        report_ctlr = 'ckanext.report.controllers:ReportController'
        map.connect('reports', '/report', controller=report_ctlr, action='index')
        map.redirect('/data/reports', '/report')
        map.connect('report', '/report/:report_name', controller=report_ctlr, action='view')
        map.connect('report-org', '/report/:report_name/:organization', controller=report_ctlr, action='view')

        # Redirects for changed report locations May 2014
        map.redirect('/reports/qa/organisation/broken_resource_links',
                     '/report/broken-links')
        map.redirect('/reports/qa/organisation/broken_resource_links/:organization',
                     '/report/broken-links/:organization')
        map.redirect('/reports/resources/:organization',
                     '/report/publisher-resources/:organization')
        # openness done
        map.redirect('/reports/qa',
                     '/report')
        map.redirect('/reports/qa/organisation/:organization',
                     '/report/openness/:organization')
        map.redirect('/api/2/util/qa/{a:.*}',
                     '/report')
        map.redirect('/reports/qa/organisation/scores',
                     '/report/openness')
        map.redirect('/reports/qa/organisation/scores/:organization',
                     '/report/openness/:organization')
        # commitments - keep in existing controller
        #   /data/reports/commitments
        #   /data/reports/commitments/cabinet-office
        # site usage - keep as it is
        # Other misc:
        #   /publisher/report_publishers_and_users
        #   /publisher/report_groups_without_admins
        #   /publisher/report_users_not_assigned_to_groups

        return map


class ThemePlugin(p.SingletonPlugin):
    '''
    DGU Visual Theme for a CKAN install embedded in dgu/Drupal.
    '''
    p.implements(p.IConfigurer)
    p.implements(p.IRoutes, inherit=True)
    p.implements(p.ITemplateHelpers, inherit=True)

    from ckan.lib.base import h, BaseController
    # [Monkey patch] Replace h.linked_user with a version to hide usernames
    h.linked_user = dgu_linked_user
    # [Monkey patch] Replace BaseController.__after__ to allow us to add more
    # headers for caching
    BaseController.__after__ = after
    # [Monkey patch] Replace TemplateController.view since it isn't used
    # in normal use. Hack attempts sometimes get through to it though
    # and there is no need to attempt to barf on their unicode characters
    from ckan.controllers.template import TemplateController
    TemplateController.view = not_found
    # [Monkey patch] Stop autodetecting package-type by the URL since it seems
    # to think /data/search should search for 'search' packages!
    from ckan.controllers.package import PackageController
    PackageController._guess_package_type = _guess_package_type
    # [Monkey patch] We have complicated UKLP license info, so need a custom
    # isopen method
    from ckan import model
    from lib import helpers as dgu_helpers
    model.Package._isopen = model.Package.isopen
    model.Package.isopen = dgu_helpers.isopen

    def update_config(self, config):
        toolkit.add_template_directory(config, 'theme/templates')
        toolkit.add_public_directory(config, 'theme/public')

        # Shared assets may be configured to be elsewhere on disk,
        # so in that case let the user configure apache/nginx to serve
        # them manually. But for developers, the shared assets will
        # simply next door to this repo, dgu.shared_assets_timestamp_path
        # won't be set, so as a convenience we get paster to serve them.
        if not config.get('dgu.shared_assets_timestamp_path'):
            toolkit.add_public_directory(config, '../../../shared_pops_assets')

    def get_helpers(self):
        """
        A dictionary of extra helpers that will be available to provide
        dgu specific helpers to the templates.  We may be able to override
        h.linked_user so that we don't need to monkey patch above.
        """
        from ckanext.dgu.lib import helpers
        from inspect import getmembers, isfunction

        helper_dict = {}

        functions_list = [o for o in getmembers(helpers, isfunction)]
        for name, fn in functions_list:
            if name[0] != '_':
                helper_dict[name] = fn

        return helper_dict

    def before_map(self, map):
        """
        Make "/data" the homepage.
        """
        data_controller = 'ckanext.dgu.controllers.data:DataController'
        user_controller = 'ckanext.dgu.controllers.user:UserController'
        package_controller = 'ckanext.dgu.controllers.package:PackageController'
        # map.redirect('/', '/data')
        #map.redirect('/data', '/data/search')
        #map.connect('/data', controller=data_controller, action='index')

        map.connect('/linked-data-admin', controller=data_controller, action='linked_data_admin')
        map.connect('dgu_search', '/search', controller=package_controller, action='search')
        map.connect('opsi_recent', '/recent', controller=package_controller, action='recent')
        map.connect('opsi_recent_podrocje', '/recent/{podrocje}', controller=package_controller, action='recent')
        map.connect('api_page', '/metadata-api-docs', controller=data_controller, action='api')
        map.connect('system_dashboard', '/system_dashboard', controller=data_controller, action='system_dashboard')
#        map.connect('openspending_index', '/data/openspending-report/index', controller=data_controller, action='openspending_report')
#        map.connect('openspending_read', '/data/openspending-report/{id}', controller=data_controller, action='openspending_publisher_report')
        map.connect('/data/resource_cache/{root}/{resource_id}/{filename}', controller=data_controller, action='resource_cache')
#        map.connect('/data/viz/social-investment-and-foundations', controller=data_controller, action='viz_social_investment_and_foundations')
#        map.connect('/data/viz/investment-readiness-programme', controller=data_controller, action='viz_investment_readiness_programme')
#        map.connect('/data/viz/upload', controller=data_controller, action='viz_upload')

        theme_controller = 'ckanext.dgu.controllers.theme:ThemeController'
#        map.connect('/data/themes', controller=theme_controller, action='index')
#        map.connect('/data/themes/{name}', controller=theme_controller, action='named_theme')

        #Summary statistics about the website
        map.connect('/site-statistics', controller=data_controller, action='site_statistics')

        # For test usage when Drupal is not running
        map.connect('/comment/get/{id}',
                    controller='ckanext.dgu.controllers.package:CommentProxy',
                    action='get_comments')

        # Remap the /user/me to the DGU version of the User controller
        with SubMapper(map, controller=user_controller) as m:
            m.connect('/user/me', action='me')
            m.connect('user_delete', '/user/delete/{id}', action='delete', conditions=dict(method=['GET', 'POST']))

        with SubMapper(map, controller='user') as m:
            m.connect('/user/edit', action='edit')
            m.connect('/user/edit/{id:.*}', action='edit')
            m.connect('/user/reset/{id:.*}', action='perform_reset')
            m.connect('/user/register', action='register')
            m.connect('/user/login', action='login')
            m.connect('/user/logout', action='logout')
            m.connect('/user/logged_in', action='logged_in')
            m.connect('/user/logged_out', action='logged_out')
            m.connect('/user/logged_out_redirect', action='logged_out_page')
            m.connect('/user/reset', action='request_reset')
            # NB not /data/user/me
            m.connect('/user/set_lang/{lang}', action='set_lang')
            m.connect('user_index', '/user', action='index')
            m.connect('user_datasets', '/user/{id:.*}', action='read',
                      ckan_icon='sitemap')
            m.connect('user_read', '/user/{id:.*}', action='read')
            # m.connect('/data/user', action='index')

        map.redirect('/dashboard', '/user/me')

        return map

    def after_map(self, map):
        # Delete routes to /tag since we use /data/tag and /tag is confusing to
        # have kicking around still when it uses the same template with
        # different inputs.
        delete_routes_by_path_startswtih(map, path_startswith='/tag')
        return map


class DrupalAuthPlugin(p.SingletonPlugin):
    '''Reads Drupal login cookies to log user in.'''
    p.implements(p.IMiddleware, inherit=True)

    def make_middleware(self, app, config):
        return DrupalAuthMiddleware(app, config)


class AuthApiPlugin(p.SingletonPlugin):
    '''Adds functions that work out if the user is allowed to do
    certain edits.'''

    p.implements(p.IAuthFunctions, inherit=True)

    def get_auth_functions(self):
        return {
                'group_update': opsi_group_update,
                'package_change_state': opsi_package_change_state,
                'package_create': opsi_package_create,
                'package_show': opsi_package_show,
                'resource_create': opsi_resource_create,
                'resource_update': opsi_resource_update,
                'resource_delete': opsi_resource_delete,
                'package_update': opsi_package_update,
                'package_extra_fields_editable': opsi_extra_fields_editable,
                'package_delete': opsi_dataset_delete,
                'user_list': opsi_user_list,
                'user_show': opsi_user_show,
                'organization_delete': dgu_organization_delete,
                'organization_update': opsi_organization_update,
                'group_change_state': dgu_group_change_state,
                'bulk_publish': opsi_bulk_publish,
                'user_delete': opsi_user_delete
               }


class DguForm(p.SingletonPlugin):
    # NB the actual form (IDatasetForm) is in forms/dataset_form.py

    p.implements(p.IRoutes, inherit=True)
    p.implements(p.IConfigurable)

    # IConfigurable
    def configure(self, config):
        from ckanext.dgu.model.package_likes import setup as model_setup
        model_setup()

    # IRoutes

    def before_map(self, map):
        dgu_package_controller = 'ckanext.dgu.controllers.package:PackageController'
        map.connect('dataset_new', '/dataset/new', controller=dgu_package_controller, action='new')
        map.connect('dataset_edit', '/dataset/edit/{id}', controller=dgu_package_controller, action='edit')
        map.connect('/dataset/delete/{id}', controller=dgu_package_controller, action='delete')
        map.connect('/dataset/change_status/{id}', controller=dgu_package_controller, action='change_status')
        map.connect('/dataset/report_opsi_issue/{id}', controller=dgu_package_controller, action='report_opsi_issue')
        map.connect('dataset_history', '/dataset/history/{id}', controller=dgu_package_controller, action='history')
        map.connect('/dataset/{id}.{format}', controller=dgu_package_controller, action='read')
        map.connect('/dataset/{id}', controller=dgu_package_controller, action='read')
        map.connect('/dataset/{id}/resource/{resource_id}', controller=dgu_package_controller, action='resource_read')
        map.connect('/dataset/{id}/resource/{resource_id}/download/{filename}', controller=dgu_package_controller, action='opsi_resource_download')
        map.connect('/dataset/sop_retriever/{id}', controller=dgu_package_controller, action='get_sop_name')
        map.connect('/dataset/increase_likes/{id}', controller=dgu_package_controller, action='increase_package_likes')
        return map


class PublisherPlugin(p.SingletonPlugin):

    p.implements(p.IRoutes, inherit=True)
    p.implements(p.ISession, inherit=True)
    p.implements(IReport)

    def before_map(self, map):
        map.redirect('/organization/{url:.*}', '/publisher/{url}')
        with SubMapper(map, controller='ckanext.dgu.controllers.publisher:PublisherController') as m:
            m.connect('publisher_index',
                     '/publisher', action='index')
            #m.connect('publisher_edit',
            #         '/publisher/edit/:id', action='edit')
            m.connect('publisher_apply',
                     '/publisher/apply/:id', action='apply')
            m.connect('publisher_apply_empty',
                     '/publisher/apply', action='apply')
            m.connect('publisher_requests',
                     '/publisher/users/requests', action='publisher_requests')
            m.connect('publisher_request',
                     '/publisher/users/request/:token', action='publisher_request')
            m.connect('publisher_request_decision',
                     '/publisher/users/request/:token/:decision',
                     action='publisher_request')
            m.connect('publisher_users',
                    '/publisher/users/:id', action='users')
            m.connect('publisher_new',
                      '/publisher/new', action='new')
            m.connect('/publisher/report_groups_without_admins',
                      action='report_groups_without_admins')
            m.connect('/publisher/report_publishers_and_users',
                      action='report_publishers_and_users')
            m.connect('/publisher/report_users',
                      action='report_users')
            m.connect('/publisher/report_users_not_assigned_to_groups',
                      action='report_users_not_assigned_to_groups')
            m.connect('publisher_read',
                      '/publisher/:id',
                      action='read')
        
        dgu_api_controller = 'ckanext.dgu.controllers.api:DguApiController'
        map.connect('/api/list_incomplete_packages', controller=dgu_api_controller, action='list_incomplete_packages')
        map.connect('/api/list_incomplete_resources', controller=dgu_api_controller, action='list_incomplete_resources')

        return map

    def after_map(self, map):
        if is_plugin_enabled('issues'):
            delete_routes_by_name(map, 'issues_for_organization')
            with SubMapper(map, controller='ckanext.issues.controller:IssueController') as m:
                m.connect('issues_for_organization', '/publisher/:org_id/issues', action='issues_for_organization')
        return map

    def update_config(self, config):
        # set the auth profile to use the publisher based auth
        config['ckan.auth.profile'] = 'publisher'

        # same for the harvesting auth profile
        config['ckan.harvest.auth.profile'] = 'publisher'

    # IReport

    def register_reports(self):
        """Register details of an extension's reports"""
        from ckanext.dgu.lib import reports
        return [#reports.nii_report_info,
                reports.publisher_activity_report_info,
                reports.publisher_resources_info,
                reports.unpublished_report_info,
                reports.draft_report_info,
                reports.datasets_without_resources_info,
                #reports.app_dataset_theme_report_info,
                #reports.app_dataset_report_info,
                reports.admin_editor_info,
                reports.licence_report_info,
                reports.completeness_report_info
                #reports.la_schemas_info,
                #reports.pdf_datasets_report_info,
                #reports.html_datasets_report_info,
                ]


class InventoryPlugin(p.SingletonPlugin):

    p.implements(p.IRoutes, inherit=True)
    p.implements(p.IConfigurer)
    p.implements(p.ISession, inherit=True)

    def before_commit(self, session):
        pass

    def before_map(self, map):
#        inv_ctlr = 'ckanext.dgu.controllers.inventory:InventoryController'
#        map.connect('/unpublished/edit-item/:id',
#                    controller=inv_ctlr, action='edit_item')
#        map.connect('unpublished_edit', '/unpublished/:id/edit',
#                    controller=inv_ctlr, action='edit')
#        map.connect('/unpublished/:id/edit/download',
#                    controller=inv_ctlr, action='download')
#        map.connect('/unpublished/:id/edit/template',
#                    controller=inv_ctlr, action='template')
#        map.connect('/unpublished/:id/edit/upload',
#                    controller=inv_ctlr, action='upload')
#        map.connect('/unpublished/:id/edit/upload_complete',
#                    controller=inv_ctlr, action='upload_complete')
#        map.connect('/unpublished/:id/edit/upload/:upload_id',
#                    controller=inv_ctlr, action='upload_status')

        return map

    def after_map(self, map):
        return map

    def update_config(self, config):
        pass


class SearchPlugin(p.SingletonPlugin):
    """
    DGU-specific searching.

    One thing DGU specific about the search is that DGU facets on
    whether a dataset's license_id is OGL (Open Government License) or not.
    Since this is calculable from the license_id, but is not a facet over the
    whole set of possible license_id values (ie. 'ukcrown', 'other' etc. should
    all be grouped together under the 'non-ogl' facet), we index on a field
    that doesn't exist on the dataset itself.  See `SearchPlugin.before_index`.

    Another thing that DGU does differently is that it cleans up the resource
    formats prior to indexing.

    A further thing that DGU does differently is to index the group title, as
    well as the group name.
    """

    p.implements(p.IPackageController, inherit=True)

    def read(self, entity):
        pass

    def create(self, entity):
        pass

    def edit(self, entity):
        pass

    def authz_add_role(self, object_role):
        pass

    def authz_remove_role(self, object_role):
        pass

    def delete(self, entity):
        pass

    def before_search(self, search_params):
        """
        Modify the search query.
        """
        # Set the 'qf' (queryfield) parameter to a fixed list of boosted solr fields
        # tuned for DGU. If a dismax query is run, then these will be the fields that are searched
        # within.
        search_params['qf'] = 'title^4 name^3 notes^2 text group_titles^0.3 extras_harvest_document_content^0.2'
        # boost NII datasets. used trial and error to get reasonable mix of useful NII ones
        # on and relevant non-NII ones for /data/search?q=road and /data/search?q=crime
        search_params['bf'] = 'core_dataset^20'

        # ignore dataset_type:dataset which CKAN2 adds in - we dont use
        # dataset_type and it mucks up spatial search
        if search_params.get('fq'):
            search_params['fq'] = search_params['fq'].replace('+dataset_type:dataset', '')

        # If the user does not specify a "sort by" method manually,
        # then it defaults here (and the UI has to have the same logic)
        # NB The UI has to be kept in step with this logic:
        #    ckanext/dgu/theme/templates/package/search.html
        order_by = search_params.get('sort')
        bbox = search_params.get('extras', {}).get('ext_bbox')
        search_params_apart_from_bbox = search_params.get('q', '') + search_params.get('fq', '')
        sort_by_location_enabled = bool(bbox and not search_params_apart_from_bbox)

        if order_by in (None, 'spatial desc') and sort_by_location_enabled:
            search_params['sort'] = 'spatial desc'
            # This sort parameter is picked up by ckanext-spatial
        elif order_by in (None, 'rank'):
            # Score = SOLR rank = relevancy = how well q keyword matches the
            # SOLR record's text content.
            # Add popularity into default ranking - this kicks when there is
            # no keyword, so no rank. Leave name there, in case no popularity
            # scores have been loaded.
            search_params['sort'] = 'score desc, popularity desc, name asc'

        return search_params

    def after_search(self, search_results, search_params):
        return search_results

    def before_view(self, pkg_dict):
        return pkg_dict

    def before_index(self, pkg_dict):
        from pylons import config
        """
        Dynamically creates a license_id-is-ogl field to index on, and clean
        up resource formats prior to indexing.
        """
        log.info('Indexing: %s', pkg_dict['name'])
        SearchIndexing.clean_title_string(pkg_dict)
        SearchIndexing.add_field__is_ogl(pkg_dict)
        SearchIndexing.resource_format_cleanup(pkg_dict)
        SearchIndexing.add_field__group_titles(pkg_dict)
        SearchIndexing.add_field__publisher(pkg_dict)
        if is_plugin_enabled('harvest'):
            SearchIndexing.add_field__harvest_document(pkg_dict)
        SearchIndexing.add_field__openness(pkg_dict)
        SearchIndexing.add_popularity(pkg_dict)
        SearchIndexing.add_filter_data_defaults(pkg_dict)
        SearchIndexing.add_field__group_abbreviation(pkg_dict)
        SearchIndexing.add_inventory(pkg_dict)
        SearchIndexing.add_api_flag(pkg_dict)
        SearchIndexing.add_theme(pkg_dict)
        if is_plugin_enabled('dgu_schema'):
            SearchIndexing.add_schema(pkg_dict)
            
        SearchIndexing.add_podrocje(pkg_dict)       #pkg_dict['all_podrocje'] = list(all_podrocje)
        SearchIndexing.add_opsi_tip(pkg_dict)
        #SearchIndexing.add_dataset_status(pkg_dict)
        SearchIndexing.add_datastore_active(pkg_dict)
        SearchIndexing.add_licence(pkg_dict)
        SearchIndexing.add_open_flag(pkg_dict)
        #SearchIndexing.add_collections(pkg_dict)

        #correct indexed urls if ckan is proxied
        rewrite_url = config.get('solr_proxy_rewrite', '')
        site_url = config.get('ckan.site_url', '')
        if rewrite_url and site_url:
            log.debug("Rewriting urls in indexer: "+str(rewrite_url)+ " for "+str(site_url))
            if 'res_url' in pkg_dict:
                for i in range(len(pkg_dict['res_url'])):
                    if pkg_dict['res_url'][i].startswith(rewrite_url):
                        pkg_dict['res_url'][i] = pkg_dict['res_url'][i].replace(rewrite_url, site_url, 1)
            if 'data_dict' in pkg_dict:
                pkg_dict['data_dict'] = pkg_dict['data_dict'].replace(rewrite_url, site_url)
            if 'validated_data_dict' in pkg_dict:
                pkg_dict['validated_data_dict'] = pkg_dict['validated_data_dict'].replace(rewrite_url, site_url)

        return pkg_dict

class ApiPlugin(p.SingletonPlugin):
    '''DGU-specific API'''
    p.implements(p.IRoutes, inherit=True)
    p.implements(p.IActions)

    def before_map(self, map):
        api_controller = 'ckanext.dgu.controllers.api:DguApiController'
        map.connect('/api/util/latest-datasets', controller=api_controller, action='latest_datasets')
        map.connect('/api/util/dataset-count', controller=api_controller, action='dataset_count')
        map.connect('/api/util/revisions', controller=api_controller, action='revisions')
        
        with SubMapper(map, controller='ckanext.dgu.controllers.api:PatchedApiController', path_prefix='/api{ver:/3|}',
                       ver='/3', logic_function='resource_create') as m:
            m.connect('/action/resource_create', action='action', conditions=dict(method=['POST']))
#        map.connect('/api/util/latest-unpublished', controller=api_controller, action='latest_unpublished')
#        map.connect('/api/util/popular-unpublished', controller=api_controller, action='popular_unpublished')

        with SubMapper(map, controller='ckanext.dgu.controllers.api:DguApiController') as m:
            m.connect('/api/util/tag/autocomplete_all', action='tag_autocomplete_all')

        return map

    def get_actions(self):
        from ckanext.dgu.logic.action.get import package_show, publisher_show, user_show_no_limit, package_autocomplete, tag_autocomplete_all
        from ckanext.dgu.logic.action.create import opsi_action_resource_create
        from ckanext.dgu.logic.action.update import opsi_action_resource_update, opsi_action_bulk_publish
        from ckanext.dgu.logic.action.create import opsi_datapusher_submit
        from ckanext.dgu.logic.action.delete import opsi_action_resource_delete
        return {
            'package_show': package_show,
            'package_autocomplete': package_autocomplete,
            'publisher_show': publisher_show,
            'user_show_no_limit': user_show_no_limit,
            'resource_create': opsi_action_resource_create,
            'resource_update': opsi_action_resource_update,
            'resource_delete': opsi_action_resource_delete,
            'tag_autocomplete_all': tag_autocomplete_all,
            'quick_datapusher_submit': opsi_datapusher_submit,
            'bulk_publish': opsi_action_bulk_publish
        }


class SiteIsDownPlugin(p.SingletonPlugin):
    '''"Site is down for maintenance" message shown for all requests - better
    than an error while there is maintenance.'''
    p.implements(p.IMiddleware, inherit=True)

    def make_middleware(self, app, config):
        return SiteDownMiddleware(app, config)

class DguSpatialPlugin(p.SingletonPlugin):
    """
    DGU-specific bits to do with spatial that are not in ckanext-spatial
    """
    p.implements(p.IDomainObjectModification, inherit=True)

    # IDomainObjectModification

    def notify(self, entity, operation=None):
        from ckan import model
        if not isinstance(entity, model.Package):
            return
        pkg = entity
        if not p.toolkit.asbool(pkg.extras.get('UKLP')):
            return

        log.debug('Notified of UKLP package event: %s %s', pkg.name, operation)

        # avoid infinite loop
        run_gemini_postprocess = \
            self._is_it_sufficient_change_to_run_gemini_postprocess(
                entity, operation)
        if not run_gemini_postprocess:
            return

        log.debug('Creating gemini post-process task: %s', pkg.name)
        gemini_postprocess_tasks.create_package_task(entity, 'priority')

    # similar to ckanext-archiver's _is_it_sufficient_change_to_run_archiver
    def _is_it_sufficient_change_to_run_gemini_postprocess(self, package,
                                                           operation):
        ''' Returns True if it is a new dataset or there are resources that
        have been added or URL changed in this revision.
        '''
        from ckan import model
        if operation == 'new':
            log.debug('New package - will process')
            # even if it has no resources, QA needs to show 0 stars against it
            return True
        elif operation == 'deleted':
            log.debug('Deleted package - won\'t process')
            return False
        # therefore operation=changed

        # check to see if resources are added or URL changed

        # look for the latest revision
        rev_list = package.all_related_revisions
        if not rev_list:
            log.debug('No sign of previous revisions - will process')
            return True
        # I am not confident we can rely on the info about the current
        # revision, because we are still in the 'before_commit' stage. So
        # simply ignore that if it's returned.
        if rev_list[0][0].id == model.Session.revision.id:
            rev_list = rev_list[1:]
        if not rev_list:
            log.warn('No sign of previous revisions - will process')
            return True
        previous_revision = rev_list[0][0]
        log.debug('Comparing with revision: %s %s',
                  previous_revision.timestamp, previous_revision.id)

        # get the package as it was at that previous revision
        context = {'model': model, 'session': model.Session,
                   #'user': c.user or c.author,
                   'ignore_auth': True,
                   'revision_id': previous_revision.id}
        data_dict = {'id': package.id}
        try:
            old_pkg_dict = p.toolkit.get_action('package_show')(
                context, data_dict)
        except p.toolkit.NotFound:
            log.warn('No sign of previous package - will process anyway')
            return True

        # have any resources been added?
        old_resources = dict((res['id'], res)
                             for res in old_pkg_dict['resources'])
        old_res_ids = set(old_resources.keys())
        new_res_ids = set((res.id for res in package.resources))
        added_res_ids = new_res_ids - old_res_ids
        if added_res_ids:
            log.debug('Added resources - will process. res_ids=%r',
                      added_res_ids)
            return True

        # have any resource urls changed?
        for res in package.resources:
            old_res_url = old_resources[res.id]['url']
            if old_res_url != res.url:
                log.debug('Resource url changed - will process. '
                          'id=%s pos=%s url="%s"->"%s"',
                          res.id[:4], res.position, old_res_url, res.url)
                return True
            log.debug('Resource unchanged. pos=%s id=%s',
                      res.position, res.id[:4])

        log.debug('No new or changed resources - won\'t process')
        return False
    
