from logging import getLogger
import re
import string
import json

from paste.deploy.converters import asbool

from ckan.model.group import Group
from ckan import model
from ckan.lib import helpers
from ckanext.dgu.lib import helpers as dgu_helpers
from ckanext.dgu.plugins_toolkit import ObjectNotFound

log = getLogger(__name__)

class SearchIndexing(object):
    '''Functions that edit the package dictionary fields to affect the way it
    gets indexed in Solr.'''

    @classmethod
    def add_popularity(cls, pkg_dict):
        '''Adds the views field from the ga-report plugin, if it is installed'''
        from pylons import config

        score = 0
        
        cutoff = config.get('ckan.tracking_cutoff')
        if cutoff:
            cutoff = int(cutoff)
        if not cutoff:
            cutoff = 1000

        if 'ga-report' in config.get('ckan.plugins'):
            from ckanext.ga_report.ga_model import get_score_for_dataset
            score += get_score_for_dataset(pkg_dict['name'])
        else:
            #Remove ga-report from CKAN plugins
            tracking_summary = pkg_dict.get('tracking_summary', None)
            if tracking_summary == None:
                tracking_summary = model.TrackingSummary.get_for_package(pkg_dict['id'])
            if tracking_summary:
                score = tracking_summary['recent']*cutoff + tracking_summary['total']

        pkg_dict['popularity'] = score
        log.debug('Popularity: %s', pkg_dict['popularity'])

    @classmethod
    def add_filter_data_defaults(cls, pkg_dict):
        pkg_dict['high_value_data'] = pkg_dict.get('high_value_data', False)
        pkg_dict['municipal'] = pkg_dict.get('municipal', False)
        pkg_dict['bulk_download'] = pkg_dict.get('bulk_download', False)

    @classmethod
    def add_api_flag(cls, pkg_dict):
        pkg_dict['api'] = 'API' in [p.upper() for p in pkg_dict['res_format']]
        log.debug('API: %s', pkg_dict['api'])

    @classmethod
    def add_inventory(cls, pkg_dict):
        ''' Sets unpublished to false if not present and also states whether the item is marked
            as never being published. '''
        pkg_dict['unpublished'] = pkg_dict.get('unpublished', False)
        #log.debug('Unpublished: %s', pkg_dict['unpublished'])

        pkg_dict['core_dataset'] = pkg_dict.get('core-dataset', False)
        #log.debug('NII: %s', pkg_dict['core_dataset'])

        # We also need to mark whether it is restricted (as in it will never be
        # released).
        pkg_dict['publish_restricted'] = pkg_dict.get('publish-restricted', False)
        #log.debug('Publish restricted: %s', pkg_dict['publish_restricted'])

    @classmethod
    def add_collections(cls, pkg_dict):
        pkg_dict['collections'] = []

        if asbool(pkg_dict.get('core-dataset', False)):
            pkg_dict['collections'].append('National Information Infrastructure')
            
        #if asbool(pkg_dict.get('podrocje', True)):
        #    pkg_dict['collections'].append('podrocje')
            
        #if dgu_helpers.is_dataset_organogram(pkg_dict):
        #    pkg_dict['collections'].append('Organograms')

        # maybe put this in a property?
        #if 'API' in [p.upper() for p in pkg_dict['res_format']]:  # what about wfs?
        #    pkg_dict['collections'].append('API access')

        log.debug('Collections: %s', pkg_dict['collections'])

    @classmethod
    def add_field__is_ogl(cls, pkg_dict):
        '''Adds the license_id-is-ogl field.'''
        if 'license_id-is-ogl' not in pkg_dict:
            is_ogl = cls._is_ogl(pkg_dict)
            pkg_dict['license_id-is-ogl'] = is_ogl
        try:
            if asbool(pkg_dict.get('unpublished', False)):
                pkg_dict['license_id-is-ogl'] = 'unpublished'
        except ValueError:
            pass

    @classmethod
    def _is_ogl(cls, pkg_dict):
        """
        Returns true iff the represented dataset has an OGL license

        A dataset has an OGL license if the license_id == "uk-ogl"
        """
        return pkg_dict['license_id'] == 'uk-ogl'

    @classmethod
    def clean_title_string(cls, pkg_dict):
        ''' Removes leading spaces from the title_string that is used for searching '''
        ts = pkg_dict.get('title_string', '').lstrip()  # strip leading whitespace
        if ts and ts[0] in string.punctuation:
            # Remove leading punctuation where we find it.
            ts = ts.replace(ts[0], '')
        pkg_dict['title_string'] = ts


    @classmethod
    def resource_format_cleanup(cls, pkg_dict):
        '''Standardises the res_format field.'''
        pkg_dict['res_format'] = [ cls._clean_format(f) for f in pkg_dict.get('res_format', []) ]

    _disallowed_characters = re.compile(r'[^a-zA-Z /+]')
    @classmethod
    def _clean_format(cls, format_string):
        if isinstance(format_string, basestring):
            matched_format = helpers.resource_formats().get(format_string.lower().strip(' .'))
            if matched_format:
                return matched_format[1]
            return re.sub(cls._disallowed_characters, '', format_string).strip()
        else:
            return format_string

    @classmethod
    def add_field__group_titles(cls, pkg_dict):
        '''Adds the group titles.'''
        groups = [Group.get(g) for g in pkg_dict['groups']]

        # Group titles
        if not pkg_dict.has_key('organization_titles'):
            pkg_dict['organization_titles'] = [g.title for g in groups]
        else:
            log.warning('Unable to add "organization_titles" to index, as the datadict '
                        'already contains a key of that name')

    @classmethod
    def add_field__group_abbreviation(cls, pkg_dict):
        '''Adds any group abbreviation '''
        abbr = None

        g = model.Group.get(pkg_dict['organization'])
        if not g:
            log.error("Package %s does not belong to an organization" % pkg_dict['name'])
            return

        try:
            abbr = g.extras.get('abbreviation')
        except:
            raise

        if abbr:
            pkg_dict['group_abbreviation'] = abbr
            #log.debug('Abbreviations: %s', abbr)

    @classmethod
    def add_field__publisher(cls, pkg_dict):
        '''Adds the 'publisher' based on group.'''
        import ckan.model as model

        publisher = model.Group.get(pkg_dict.get('organization'))
        if not publisher:
            log.warning('Dataset %s doesn\'t seem to have a publisher!  '
                        'Unable to add publisher to index.',
                        pkg_dict['name'])
            return pkg_dict

        # Publisher names
        if not pkg_dict.has_key('publisher'):
            pkg_dict['publisher'] = publisher.name
            log.debug(u"Publisher: %s", publisher.name)
        else:
            log.warning('Unable to add "publisher" to index, as the datadict '
                        'already contains a key of that name')

        # Ancestry of publishers
        ancestors = []
        while(publisher is not None):
            ancestors.append(publisher)
            parent_publishers = publisher.get_parent_groups('organization')
            if len(parent_publishers) == 0:
                publisher = None
            else:
                if len(parent_publishers) > 1:
                    log.warning('Publisher %s has more than one parent publisher. '
                                'Ignoring all but the first. %s',
                                publisher, repr(parent_publishers))
                publisher = parent_publishers[0]


        if not pkg_dict.has_key('parent_publishers'):
            pkg_dict['parent_publishers'] = [ p.name for p in ancestors ]
        else:
            log.warning('Unable to add "parent_publishers" to index, as the datadict '
                        'already contains a key of that name. '
                        'Package: %s Parent_publishers: %r', \
                        pkg_dict['name'], pkg_dict['parent_publishers'])

    @classmethod
    def add_field__harvest_document(cls, pkg_dict):
        '''Index a harvested dataset\'s XML content
           (Given a low priority when searching)'''
        if pkg_dict.get('UKLP', '') == 'True':
            import ckan
            from ckanext.dgu.plugins_toolkit import get_action

            context = {'model': ckan.model,
                       'session': ckan.model.Session,
                       'ignore_auth': True}

            data_dict = {'id': pkg_dict.get('harvest_object_id', '')}

            try:
                harvest_object = get_action('harvest_object_show')(context, data_dict)
                pkg_dict['extras_harvest_document_content'] = harvest_object.get('content', '')
            except ObjectNotFound:
                log.warning('Unable to find harvest object "%s" '
                            'referenced by dataset "%s"',
                            data_dict['id'], pkg_dict['id'])

    @classmethod
    def add_field__openness(cls, pkg_dict):
        '''Add the openness score (stars) to the search index'''
        archival = pkg_dict.get('archiver')
        if not archival or dgu_helpers.is_string(archival):
            log.warning('No Archiver info for package %s, is string: %s', pkg_dict['name'], str(dgu_helpers.is_string(archival)))
            return
        qa = pkg_dict.get('qa')
        if not qa:
            log.warning('No QA info for package %s', pkg_dict['name'])
            return
        pkg_dict['openness_score'] = qa.get('openness_score')
        log.debug('Openness score: %s', pkg_dict['openness_score'])

        if not hasattr(cls, 'broken_links_map'):
            cls.broken_links_map = {
                    True: 'Pokvarjeni',
                    False: 'OK',
                    None: 'Nepreverjeni'
                    }
        pkg_dict['broken_links'] = cls.broken_links_map[archival.get('is_broken')]
        log.debug('Broken links: %s', pkg_dict['broken_links'])

    @classmethod
    def add_theme(cls, pkg_dict):
        # Extract all primary and secondary themes into the 'all_themes' field
        all_themes = set()
        if pkg_dict.get('theme-primary'):
            all_themes.add(pkg_dict.get('theme-primary', ''))
        try:
            secondary_themes = json.loads(pkg_dict.get('theme-secondary', '[]'))
            for theme in secondary_themes:
                if theme:
                    all_themes.add(theme)
        except ValueError, e:
            log.error('Could not parse secondary themes: %s %r',
                      pkg_dict['name'], pkg_dict.get('theme-secondary'))
        pkg_dict['all_themes'] = list(all_themes)
        log.debug('Themes: %s', ' '.join(all_themes))

    @classmethod
    def add_open_flag(cls, pkg_dict):
        '''Add if the package represents open data to the search index'''
        
        pkg_dict['open_data'] = dgu_helpers.is_open_data(pkg_dict)
        log.debug('Open data: %s', pkg_dict['open_data'])
        
    @classmethod
    def add_opsi_tip(cls, pkg_dict):
        # Extract all opsi_tip_naziv into the 'all_opsi_tip' field
        
        pkg_dict['all_opsi_tip'] = []
        opsi_tip = pkg_dict.get('opsi_tip_naziv', '')
        pkg_dict['all_opsi_tip'].append(opsi_tip)
        log.debug('Tip organizacije: %s', ' '.join(opsi_tip))
        
    @classmethod
    def add_podrocje(cls, pkg_dict):
        # Extract all podrocja into the 'all_podrocje' field
        #all_podrocje = set()
        #if pkg_dict.get('podrocje'):
        #    all_podrocje.add(pkg_dict.get('podrocje', ''))
        #pkg_dict['all_podrocje'] = list(all_podrocje)

        #if asbool(pkg_dict.get('podrocja', True)):
        
        pkg_dict['all_podrocje'] = []
        podrocje = pkg_dict.get('podrocje', '')
        #podrocje = helpers.get_pkg_dict_extra(pkg_dict, 'podrocje', default=None)
        #if podrocje:
        #    if isinstance(podrocje, types.ListType):
        #        podrocje = ', '.join(podrocje)
        #    if isinstance(podrocje, types.StringType):
        #        podrocje = unicode(podrocje)
        pkg_dict['all_podrocje'].append(podrocje)

        #Vrne prazen string
        log.debug('Podrocje: %s', ' '.join(podrocje))

    @classmethod
    def add_datastore_active(cls, pkg_dict):
        # Add information if package has data in datastore with an artificial format
        for r in pkg_dict.get('resources', []):
            if r.get('datastore_active', False):
                pkg_dict['res_format'] = pkg_dict.get('res_format', []).append('Prog. vmesnik')
                return
        try:
            data_dict = json.loads(pkg_dict.get('data_dict',''))
            for r in data_dict.get('resources', []):
                if r.get('datastore_active', False):
                    pkg_dict.setdefault('res_format', []).append('Prog. vmesnik')
                    return
        except ValueError:  # includes JSONDecodeError
            log.warning('data_dict of package could not be decoded, datastore_active flag may not be set correctly')

    @classmethod
    def add_licence(cls, pkg_dict):
        # Extract all licence into the 'all_licence' field
        
        pkg_dict['all_licence'] = []
        
        #licence=""
        #try:
        #    licence = pkg_dict.get('licence', '').encode("ascii")
        #except UnicodeDecodeError:
        #    licence = pkg_dict.get('licence', '').encode("utf-8")
        licence = pkg_dict.get('licence', '')
            
        #podrocje = helpers.get_pkg_dict_extra(pkg_dict, 'podrocje', default=None)
        #if podrocje:
        #    if isinstance(podrocje, types.ListType):
        #        podrocje = ', '.join(podrocje)
        #    if isinstance(podrocje, types.StringType):
        #        podrocje = unicode(podrocje)
        
        #Vrne prazen string
        log.debug('Licenca: %s', ' '.join(licence))
        
        #Harvestano licenco si zapomni le, ce je ena izmed uradnih licenc
        #Ce podatkovna zbirka ni harvestana, si licenco vedno zapomni
        
        zbirka_licenc=[]
        for id_option, licenca_option in dgu_helpers.license_choices(pkg_dict):
            zbirka_licenc.append(licenca_option)
        max_licence_length = -1
        if dgu_helpers.was_dataset_harvested(pkg_dict):
            log.debug('This dataset was harvested')
            
            if licence in zbirka_licenc:
                log.debug('Licence can be found in %s', ' '.join(zbirka_licenc))
                pkg_dict['all_licence'].append(licence)
            #Indeksiram lahko se, ce je besedilo licence dovolj kratko (par besed)
            elif len(licence) <= max_licence_length:
                log.debug('Licence indexed even though the dataset was harvested, as licence text was shorter than %s characters', ' '.join(str(max_licence_length)))
                pkg_dict['all_licence'].append(licence)
            else:
                log.debug('Licence not indexed due to the dataset being harvested and licence text being longer than %s characters', ' '.join(str(max_licence_length)))
        else:
            pkg_dict['all_licence'].append(licence)
            log.debug('Licence indexed due to dataset not being harvested.')

    @classmethod
    def add_schema(cls, pkg_dict):
        from ckanext.dgu.model.schema_codelist import Schema, Codelist
        try:
            schema_ids = json.loads(pkg_dict.get('schema') or '[]')
        except ValueError:
            log.error('Not valid JSON in schema field: %s %r',
                      pkg_dict['name'], pkg_dict.get('schema'))
            schema_ids = None
        schemas = []
        for schema_id in schema_ids:
            try:
                schemas.append(Schema.get(schema_id).title)
            except AttributeError, e:
                log.error('Invalid schema_id: %r %s', schema_id, e)
        pkg_dict['schema_multi'] = schemas
        #log.debug('Schema: %s', ' '.join(schemas))

        try:
            codelist_ids = json.loads(pkg_dict.get('codelist') or '[]')
        except ValueError:
            log.error('Not valid JSON in codelists field: %s %r',
                      pkg_dict['name'], pkg_dict.get('codelist'))
            codelists = None
        codelists = []
        for codelist_id in codelist_ids:
            codelists.append(Codelist.get(codelist_id).title)
        pkg_dict['codelist_multi'] = codelists
        #log.debug('Code lists: %s', ' '.join(codelists))

